import anndata
import numpy as np
import os
import pandas
import zipfile
import scipy.io
from sfaira.data import DatasetBase


class Dataset(DatasetBase):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.download_url_data = \
            "https://www.brainimmuneatlas.org/data_files/toDownload/filtered_gene_bc_matrices_mex_WT_fullAggr.zip"
        self.download_url_meta = \
            "https://www.brainimmuneatlas.org/data_files/toDownload/annot_fullAggr.csv"

        self.assay_sc = "10x 3' v2"
        self.author = "Hove"
        self.disease = "healthy"
        self.doi_journal = "10.1038/s41593-019-0393-4"
        self.layer_counts = "X"
        self.organism = "Mus musculus"
        self.primary_data = True
        self.sample_source = "primary_tissue"
        self.state_exact = "healthy"
        self.year = 2019

        self.bio_sample_obs_key = "sample"
        self.cell_type_obs_key = "cluster"
        self.organ_obs_key = "organ"

        self.feature_id_var_key = "ensembl"
        self.feature_symbol_var_key = "name"
        self.feature_type = "rna"

        self.set_dataset_id(idx=1)


def load(data_dir, **kwargs):
    sample_organ_dict = {
        "Choroid plexus": "choroid plexus",
        "Dura mater": "dura mater",
        "Enr. SDM": "brain meninx",
        "Whole brain": "brain",
    }
    fn = [
        os.path.join(data_dir, "filtered_gene_bc_matrices_mex_WT_fullAggr.zip"),
        os.path.join(data_dir, "annot_fullAggr.csv")
    ]

    with zipfile.ZipFile(fn[0]) as archive:
        x = scipy.io.mmread(archive.open('filtered_gene_bc_matrices_mex/mm10/matrix.mtx')).T.tocsr()
        adata = anndata.AnnData(x)
        var = pandas.read_csv(archive.open('filtered_gene_bc_matrices_mex/mm10/genes.tsv'), sep="\t", header=None)
        var.columns = ["ensembl", "name"]
        obs_names = pandas.read_csv(archive.open('filtered_gene_bc_matrices_mex/mm10/barcodes.tsv'),
                                    sep="\t",
                                    header=None
                                    )[0].values
    obs = pandas.read_csv(fn[1])
    obs.fillna("isnan", inplace=True)

    # Match annotation to raw data.
    obs.index = obs["cell"].values
    obs = obs.loc[[i in obs_names for i in obs.index], :]
    idx_tokeep = np.where([i in obs.index for i in obs_names])[0]
    adata = adata[idx_tokeep, :]
    obs_names = obs_names[idx_tokeep]
    idx_map = np.array([obs.index.tolist().index(i) for i in obs_names])
    adata = adata[idx_map, :]
    obs_names = obs_names[idx_map]
    obs["organ"] = [sample_organ_dict[x] for x in obs["sample"].values]

    # Assign attributes
    adata.obs_names = obs_names
    adata.var = var
    adata.obs = obs

    return adata



import logging
import sys
from typing import Optional, List, Union

import questionary
from prompt_toolkit.styles import Style  # type: ignore
from rich import print

log = logging.getLogger(__name__)

sfaira_style = Style([
    ('qmark', 'fg:#0000FF bold'),  # token in front of the question
    ('question', 'bold'),  # question text
    ('answer', 'fg:#008000 bold'),  # submitted answer text behind the question
    ('pointer', 'fg:#0000FF bold'),  # pointer used in select and checkbox prompts
    ('highlighted', 'fg:#0000FF bold'),  # pointed-at choice in select and checkbox prompts
    ('selected', 'fg:#008000'),  # style for a selected item of a checkbox
    ('separator', 'fg:#cc5454'),  # separator in lists
    ('instruction', ''),  # user instructions for select, rawselect, checkbox
    ('text', ''),  # plain text
    ('disabled', 'fg:#FF0000 italic')  # disabled choices for select and checkbox prompts
])


def sfaira_questionary(function: str,
                       question: str,
                       choices: Optional[List[str]] = None,
                       default: Optional[str] = None) -> Union[str, bool]:
    """
    Custom selection based on Questionary. Handles keyboard interrupts and default values.

    :param function: The function of questionary to call (e.g. select or text).
        See https://github.com/tmbo/questionary for all available functions.
    :param question: The question to prompt for. Should not include default values or colons.
    :param choices: List of all possible choices. Usually only relevant with 'select'.
    :param default: A set default value which will be chosen if the user does not enter anything.
    :return: The chosen answer.
    """
    answer: Optional[str] = ''
    try:
        if function == 'select':
            if default not in choices:  # type: ignore
                log.debug(f'Default value {default} is not in the set of choices!')
            answer = getattr(questionary, function)(f'{question}: ', choices=choices, style=sfaira_style).unsafe_ask()
        elif function == 'password':
            # while not answer or answer == '': # re-asking behaviour is unwanted for sfaira cli pull request submission
            answer = getattr(questionary, function)(f'{question}: ', style=sfaira_style).unsafe_ask()
        elif function == 'text':
            if not default:
                log.debug('Tried to utilize default value in questionary prompt, '
                          'but is None! Please set a default value.')
                default = ''
            answer = getattr(questionary, function)(f'{question} [{default}]: ', style=sfaira_style).unsafe_ask()
        elif function == 'confirm':
            default_value_bool = True if default == 'Yes' or default == 'yes' else False
            answer = getattr(questionary, function)(f'{question} [{default}]: ', style=sfaira_style,
                                                    default=default_value_bool).unsafe_ask()
        else:
            log.debug(f'Unsupported questionary function {function} used!')

    except KeyboardInterrupt:
        print('[bold red] Aborted!')
        sys.exit(1)
    if answer is None or answer == '':
        answer = default

    log.debug(f'User was asked the question: ||{question}|| as: {function}')
    log.debug(f'User selected {answer}')

    if isinstance(answer, str):
        answer = answer.strip('\"')
    return answer  # type: ignore



CELLTYPEMLP_TOPOLOGIES = {
    "0.0.1": {
        "model_type": "mlp",
        "input": {
            "genome": "Homo_sapiens.GRCh38.102",
            "genes": ["biotype", "protein_coding"],
        },
        "output": {
            "cl": "v2021-02-01",
            "targets": None,
        },
        "hyper_parameters": {
            "units": [],
            "activation": None,
            "use_bias": True,
            "l1_coef": 0.,
            "l2_coef": 0.,
            "kernel_initializer": 'glorot_uniform',
            "bias_initializer": 'zeros',
            "bias_regularizer": None,
            "activity_regularizer": None,
            "kernel_constraint": None,
            "bias_constraint": None
        }
    },
    "0.1.1": {
        "model_type": "mlp",
        "input": {
            "genome": "Homo_sapiens.GRCh38.102",
            "genes": ["biotype", "protein_coding"],
        },
        "output": {
            "cl": "v2021-02-01",
            "targets": None,
        },
        "hyper_parameters": {
            "units": [128],
            "activation": "selu",
            "use_bias": True,
            "l1_coef": 0.,
            "l2_coef": 0.,
            "kernel_initializer": 'lecun_normal',
            "bias_initializer": 'zeros',
            "bias_regularizer": None,
            "activity_regularizer": None,
            "kernel_constraint": None,
            "bias_constraint": None
        }
    },
    "0.1.2": {
        "model_type": "mlp",
        "input": {
            "genome": "Homo_sapiens.GRCh38.102",
            "genes": ["biotype", "protein_coding"],
        },
        "output": {
            "cl": "v2021-02-01",
            "targets": None,
        },
        "hyper_parameters": {
            "units": [256, 128],
            "activation": "selu",
            "use_bias": True,
            "l1_coef": 0.,
            "l2_coef": 0.,
            "kernel_initializer": 'lecun_normal',
            "bias_initializer": 'zeros',
            "bias_regularizer": None,
            "activity_regularizer": None,
            "kernel_constraint": None,
            "bias_constraint": None
        }
    },
    "0.1.3": {
        "model_type": "mlp",
        "input": {
            "genome": "Homo_sapiens.GRCh38.102",
            "genes": ["biotype", "protein_coding"],
        },
        "output": {
            "cl": "v2021-02-01",
            "targets": None,
        },
        "hyper_parameters": {
            "units": [512, 256, 128],
            "activation": "selu",
            "use_bias": True,
            "l1_coef": 0.,
            "l2_coef": 0.,
            "kernel_initializer": 'lecun_normal',
            "bias_initializer": 'zeros',
            "bias_regularizer": None,
            "activity_regularizer": None,
            "kernel_constraint": None,
            "bias_constraint": None
        }
    }
}

# Load versions from extension if available:
try:
    from sfaira_extension.versions.topology_versions.human.celltype import CELLTYPEMLP_TOPOLOGIES as CELLTYPEMLP_TOPOLOGIES_EXTENSION
    CELLTYPEMLP_TOPOLOGIES = {
        **CELLTYPEMLP_TOPOLOGIES,
        **CELLTYPEMLP_TOPOLOGIES_EXTENSION
    }
except ImportError:
    pass



import anndata
import gzip
import os
import pandas as pd
import tarfile
import scanpy as sc
import scipy.sparse


def ir_read_from_chain_gz(fn, sep, barcodes, id_col, locus_col, v_call_col, d_call_col, j_call_col, c_call_col,
                          productive_col, junction_col, junction_aa_col, consensus_count_col, duplicate_count_col):
    """
    Util function to read the VDJ files with custom record format

    Replaces ir.io.read_airr.
    ToDo: Can this code be moved to scirpy? This file format does not seem very rare.

    E.g. of the form d10_1126_science_abe6474::GSE156728_10X_VDJ.merge.txt.gz.
    Issues with ir.io.read_airr on this file are:

        - gzip compression
        - non-standard column names
        - selected extraction of cells into AIRR format to save time
    """
    from scirpy.io._io import from_airr_cells, AirrCell, DEFAULT_AIRR_CELL_ATTRIBUTES, DEFAULT_AIRR_FIELDS

    tab = pd.read_csv(fn, sep=sep, compression="gzip")
    airr_cells = {}
    for bc in barcodes:
        airr_cells[bc] = AirrCell(
            cell_id=bc,
            cell_attribute_fields=DEFAULT_AIRR_CELL_ATTRIBUTES,
        )
        # Get all chains for this barcode:
        tab_i = tab.loc[tab[id_col].values == bc, :]
        for i in range(tab_i.shape[0]):
            chain = {
                "productive": tab_i[productive_col].values[i],
                "locus": tab_i[locus_col].values[i],
                "v_call": tab_i[v_call_col].values[i],
                "d_call": tab_i[d_call_col].values[i],
                "j_call": tab_i[j_call_col].values[i],
                "c_call": tab_i[c_call_col].values[i],
                "junction": tab_i[junction_col].values[i],
                "junction_aa": tab_i[junction_aa_col].values[i],
                "consensus_count": tab_i[consensus_count_col].values[i],
                "duplicate_count": tab_i[duplicate_count_col].values[i],
                "sequence_id": f"{bc}_{i}",
                "sequence": None,
                "rev_comp": None,
                "sequence_alignment": None,
                "germline_alignment": None,
                "v_cigar": None,
                "d_cigar": None,
                "j_cigar": None,
            }
            airr_cells[bc].add_chain(chain)

    return from_airr_cells(airr_cells.values(), include_fields=DEFAULT_AIRR_FIELDS)


def load(data_dir, sample_fn, **kwargs):
    import scirpy as ir

    fn = os.path.join(data_dir, sample_fn + ".counts.txt.gz")
    # Some of the count matrices are in a tar archive, the rest is given as individual downloads
    if sample_fn.startswith("GSE156728_RAW/"):
        fn_tar = os.path.join(data_dir, "GSE156728_RAW.tar")
        with tarfile.open(fn_tar) as tar:
            with gzip.open(tar.extractfile(sample_fn.split("GSE156728_RAW/")[-1] + ".counts.txt.gz"), "rb") as f:
                tab = pd.read_csv(f, sep="\t", index_col=0).T
            adata = anndata.AnnData(tab)
    else:
        adata = sc.read(fn).transpose()
        adata.X = scipy.sparse.csr_matrix(adata.X)
    fn_meta = os.path.join(data_dir, "GSE156728_metadata.txt.gz")
    tab_meta = pd.read_csv(fn_meta, sep="\t", compression="gzip")
    tab_meta.index = tab_meta["cellID"].values
    del tab_meta["cellID"]
    adata.obs = tab_meta.loc[adata.obs_names, :]
    fn_vdj = os.path.join(data_dir, "GSE156728_10X_VDJ.merge.txt.gz")
    adata_tcr = ir_read_from_chain_gz(
        fn=fn_vdj,
        barcodes=adata.obs_names,
        sep="\t",
        id_col="barcode",
        locus_col="chain",
        v_call_col="v_gene",
        d_call_col="d_gene",
        j_call_col="j_gene",
        c_call_col="c_gene",
        productive_col="productive",
        junction_col="cdr3_nt",
        junction_aa_col="cdr3",
        consensus_count_col="umis",
        duplicate_count_col="reads",
    )
    ir.pp.merge_with_ir(adata, adata_tcr)
    return adata



import anndata
import gzip
import os
import pandas as pd
import scipy.io
import tarfile


def load(data_dir, **kwargs):
    fn = os.path.join(data_dir, "GSE164378_RAW.tar")
    adatas = []
    with tarfile.open(fn) as tar:
        samples = ['GSM5008737_RNA_3P', 'GSM5008738_ADT_3P']
        for sample in samples:
            with gzip.open(tar.extractfile(sample + '-matrix.mtx.gz'), 'rb') as mm:
                x = scipy.io.mmread(mm).T.tocsr()
            obs = pd.read_csv(tar.extractfile(sample + '-barcodes.tsv.gz'), compression='gzip',
                              header=None, sep='\t', index_col=0)
            obs.index.name = None
            var = pd.read_csv(tar.extractfile(sample + '-features.tsv.gz'), compression='gzip',
                              header=None, sep='\t').iloc[:, :1]
            var.columns = ['names']
            var.index = var['names'].values
            adata = anndata.AnnData(X=x, obs=obs, var=var)
            adata.var_names_make_unique()
            adatas.append(adata)
        tar.close()

    adata = adatas[0]
    protein = adatas[1]
    meta = pd.read_csv(os.path.join(data_dir, 'GSE164378_sc.meta.data_3P.csv.gz'), index_col=0)
    adata.obs = adata.obs.join(meta)
    adata.obsm['protein_expression'] = pd.DataFrame(protein.X.A, columns=protein.var_names, index=protein.obs_names)

    return adata



import anndata
import dask.array
import dask.dataframe
import numpy as np
import os
import pandas as pd
from pathlib import Path
import pickle
import scipy.sparse
from typing import List, Tuple, Union
import zarr


def _invert_permutation(p: np.ndarray):
    s = np.empty_like(p)
    s[p] = np.arange(p.size)
    return s


def _assign_perm_original_data_column(df, perm):
    if perm is not None:
        df['permutation_original_data'] = perm

    return df


def _buffered_path(path_base, path, fn):
    path_base = os.path.join(path_base, path)
    if not os.path.exists(path_base):
        os.makedirs(path_base)
    return os.path.join(path_base, fn)


def path_obs(path):
    return _buffered_path(path_base=path, path="parquet", fn="obs.parquet")


def path_var(path):
    return _buffered_path(path_base=path, path="parquet", fn="var.parquet")


def path_uns(path):
    return _buffered_path(path_base=path, path="pickle", fn="uns.pickle")


def path_x(path):
    if not os.path.exists(path):
        os.makedirs(path)
    return os.path.join(path, "zarr")


def write_dao(store: Union[str, Path], adata: anndata.AnnData, chunks: Union[bool, Tuple[int, int]],
              compression_kwargs: dict, shuffle_data: bool = False):
    """
    Writes a distributed access optimised ("dao") store of a dataset based on an AnnData instance.

    The following components are saved:
        - .X: as zarr array which can be interfaced with zarr or dask (or xarray).
        - .obs: as parquet table which can be interfaced with pandas.DataFrame (and dask.dataframe.DataFrame).
        - .var: as parquet table which can be interfaced with pandas.DataFrame (and dask.dataframe.DataFrame).
        - .uns: as a pickle to be flexible with values here.

    TODO: If layers become relevant for this store, they can be added into the zarr group.
    TODO: If obsp, varp become relevant for this store, they can be added into the zarr group.
    TODO: If obsm, varm become relevant for this store, they can be added into the zarr group.

    :param store: File name of the store (zarr group).
    :param adata: Anndata to save.
    :param chunks: Chunking of .X for zarr.
    :param compression_kwargs: Compression kwargs for zarr.
    :param shuffle_data: If True -> shuffle ordering of cells in dataset before writing store to disk
    """
    # Write numeric matrix as zarr array:
    f = zarr.open(store=path_x(store), mode="w")
    if shuffle_data:
        perm = np.arange(0, adata.X.shape[0])
        np.random.shuffle(perm)
        perm_original_data = _invert_permutation(perm)
    else:
        perm = np.arange(0, adata.X.shape[0])
        perm_original_data = None
    # If adata.X is already a dense array in memory, it can be safely written fully to a zarr array. Otherwise:
    # Create empty store and then write in dense chunks to avoid having to load entire adata.X into a dense array in
    # memory.
    if isinstance(adata.X, np.ndarray) or isinstance(adata.X, np.matrix):
        f.create_dataset("X", data=np.array(adata.X)[perm], chunks=chunks, dtype=adata.X.dtype, **compression_kwargs)
    elif isinstance(adata.X, scipy.sparse.csr_matrix):
        # Initialise empty array
        dtype = adata.X.data.dtype
        shape = adata.X.shape
        f.create_dataset("X", shape=shape, dtype=dtype, fill_value=0., chunks=chunks, **compression_kwargs)
        batch_size = 128  # Use a batch size that guarantees that the dense batch fits easily into memory.
        n_batches = shape[0] // batch_size + int(shape[0] % batch_size > 0)
        batches = [(i * batch_size, min(i * batch_size + batch_size, shape[0])) for i in range(n_batches)]
        x_permuted = adata.X[perm]
        for s, e in batches:
            f["X"][s:e, :] = np.asarray(x_permuted[s:e, :].todense(), dtype=dtype)
    else:
        raise ValueError(f"did not recognise array format {type(adata.X)}")
    # Write .uns into pickle:
    with open(path_uns(store), "wb") as f:
        # convert to dict to get rid of anndata OverloadedDict
        pickle.dump(obj=dict(adata.uns), file=f)
    # Write .obs and .var as a separate file as this can be easily interfaced with DataFrames.
    (
        adata
        .obs.iloc[perm]
        # only assign permutation_original_data column if data was shuffled
        .pipe(_assign_perm_original_data_column, perm=perm_original_data)
        # make sure all columns are dtype=str and are converted to categorical
        # this has to change if we have numeric values in obs
        # exclude 'perumtation_original_data' column from this as it's numeric
        .astype({col: str for col in adata.obs.columns if col not in ['permutation_original_data']})
        .astype({col: 'category' for col in adata.obs.columns if col not in ['permutation_original_data']})
        .to_parquet(path=path_obs(store), engine='pyarrow', compression='snappy', index=None)
    )
    adata.var.to_parquet(path=path_var(store), engine='pyarrow', compression='snappy', index=None)


def read_dao(store: Union[str, Path], use_dask: bool = True, columns: Union[None, List[str]] = None,
             obs_separate: bool = False, x_separate: bool = False) -> \
        Union[Tuple[anndata.AnnData, Union[dask.dataframe.DataFrame, pd.DataFrame]], anndata.AnnData]:
    """
    Assembles an AnnData instance based on distributed access optimised ("dao") store of a dataset.

    See write_distributed_access_optimised() for the expected format of the store.
    In particular, if use_dask is True:

        - .X is interfaced as a dask Array

    Can return representation of .obs separately, which makes sense if a HPC framework is used for this tabular format
    which is not supported by AnnData as an .obs entry.

    :param store: Path to zarr group.
    :param use_dask: Whether to use lazy dask arrays where appropriate.
    :param columns: Which columns to read into the obs copy in the output, see pandas.read_parquet().
    :param obs_separate: Whether to return .obs as a separate return value or in the returned AnnData.
    :param x_separate: Whether to return .X as a separate return value or in the returned AnnData.
    :return: Tuple of:
        - AnnData with .X as dask array.
        - obs table separately as dataframe
    """
    assert not (obs_separate and x_separate), "either request obs_separate or x_separate, or neither, but not both"
    if use_dask:
        x = dask.array.from_zarr(url=path_x(store), component="X")
    else:
        f = zarr.open(path_x(store), mode="r")
        x = f["X"]  # Select member of group.
        x = x[...]  # Load into memory.
    # Read pickle:
    with open(path_uns(store), "rb") as f:
        uns = pickle.load(file=f)
    # Read tables:
    obs = pd.read_parquet(path_obs(store), columns=columns, engine="pyarrow")
    var = pd.read_parquet(path_var(store), engine="pyarrow")
    d = {"var": var, "uns": uns}
    # Assemble AnnData without obs to save memory:
    adata = anndata.AnnData(**d, shape=x.shape)
    # Need to add these attributes after initialisation so that they are not evaluated:
    if not x_separate:
        adata.X = x
    if not obs_separate:
        adata.obs = obs
    if obs_separate:
        return adata, obs
    elif x_separate:
        return adata, x
    else:
        return adata



from sfaira.estimators.keras.base import EstimatorKeras, EstimatorKerasEmbedding, EstimatorKerasCelltype

try:
    from sfaira_extension.estimators import *  # noqa: F403
except ImportError:
    pass



import anndata
import os
import pandas as pd

from sfaira.data import DatasetBase


class Dataset(DatasetBase):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.download_url_data = "https://ftp.ncbi.nlm.nih.gov/geo/series/GSE124nnn/GSE124395/suppl/GSE124395%5FNormalhumanlivercellatlasdata%2Etxt%2Egz"
        self.download_url_meta = "https://ftp.ncbi.nlm.nih.gov/geo/series/GSE124nnn/GSE124395/suppl/GSE124395%5Fclusterpartition%2Etxt%2Egz"

        self.assay_sc = "CEL-seq2"
        self.author = "Aizarani"
        self.disease = "healthy"
        self.doi_journal = "10.1038/s41586-019-1373-2"
        self.layer_counts = "X"
        self.sample_source = "primary_tissue"
        self.organ = "liver"
        self.organism = "Homo sapiens"
        self.primary_data = True
        self.year = 2019

        self.feature_symbol_var_key = "index"
        self.feature_type = "rna"
        self.cell_type_obs_key = "CellType"

        self.set_dataset_id(idx=1)


def load(data_dir, **kwargs):
    fn = [
        os.path.join(data_dir, "GSE124395_Normalhumanlivercellatlasdata.txt.gz"),
        os.path.join(data_dir, "GSE124395_clusterpartition.txt.gz")
    ]
    adata = anndata.AnnData(pd.read_csv(fn[0], sep="\t").T)
    celltype_df = pd.read_csv(fn[1], sep=" ")
    adata = adata[[i in celltype_df.index for i in adata.obs.index]].copy()
    adata.obs["CellType"] = [str(celltype_df.loc[i]["sct@cpart"]) for i in adata.obs.index]

    return adata



from typing import Union

try:
    import sfaira_extension as sfairae
except ImportError:
    sfairae = None

from sfaira.data.dataloaders.loaders import DatasetSuperGroupLoaders
from sfaira.data.dataloaders.databases import DatasetSuperGroupDatabases
from sfaira.data.dataloaders.base.dataset_group import DatasetSuperGroup


class Universe(DatasetSuperGroup):

    def __init__(
            self,
            data_path: Union[str, None] = None,
            meta_path: Union[str, None] = None,
            cache_path: Union[str, None] = None,
            exclude_databases: bool = True,
    ):
        """
        Nested super group of data loaders, unifying data set wise data loader SuperGroup and the database
        interface SuperGroup.

        :param data_path:
        :param meta_path:
        :param cache_path:
        """
        dsgs = [
            DatasetSuperGroupLoaders(
                data_path=data_path,
                meta_path=meta_path,
                cache_path=cache_path,
            ),
        ]
        if not exclude_databases:
            dsgs.append(
                DatasetSuperGroupDatabases(
                    data_path=data_path,
                    meta_path=meta_path,
                    cache_path=cache_path,
                )
            )
        if sfairae is not None:
            dsgs.append(sfairae.data.dataloaders.loaders.DatasetSuperGroupLoaders(
                data_path=data_path,
                meta_path=meta_path,
                cache_path=cache_path,
            ))
        super().__init__(dataset_groups=dsgs)



import anndata
import numpy as np
import os
import re
import pandas as pd
import pickle
from typing import List, Union
import warnings
import time

from sfaira import settings
from sfaira.consts import AdataIdsSfaira, AdataIds, OCS
from sfaira.data import DatasetInteractive
from sfaira.estimators.keras.base import EstimatorKerasEmbedding, EstimatorKerasCelltype
from sfaira.ui.model_zoo import ModelZoo
from sfaira.versions.topologies import TopologyContainer


class UserInterface:
    """
    This class performs data set handling and coordinates estimators for the different model types.
    Example code to obtain a UMAP embedding plot of the embedding created from your data with cell-type labels::

        import sfaira
        import anndata
        import scanpy

        # initialise your sfaira instance with a model lookuptable.
        ui = sfaira.ui.UserInterface(custom_repo="/path/to/local/repo/folder/or/zenodo/repo/URL", sfaira_repo=False)
        ui.zoo_embedding.model_id = 'embedding_human-blood-ae-0.2-0.1_theislab'  # pick desired model here
        ui.zoo_celltype.model_id = 'celltype_human-blood-mlp-0.1.3-0.1_theislab'  # pick desired model here
        ui.load_data(anndata.read("/path/to/file.h5ad"), gene_symbol_col='index', gene_ens_col='gene_ids')
        ui.load_model_embedding()
        ui.load_model_celltype()
        ui.predict_all()
        adata = ui.data.adata
        scanpy.pp.neighbors(adata, use_rep="X_sfaira")
        scanpy.tl.umap(adata)
        scanpy.pl.umap(adata, color="celltypes_sfaira", show=True, save="UMAP_sfaira.png")

    """

    estimator_embedding: Union[EstimatorKerasEmbedding, None]
    estimator_celltype: Union[EstimatorKerasCelltype, None]
    zoo_embedding: Union[ModelZoo, None]
    zoo_celltype: Union[ModelZoo, None]
    data: Union[DatasetInteractive, None]
    model_lookuptable: Union[pd.DataFrame, None]
    adata_ids: AdataIds

    def __init__(
            self,
            custom_repo: Union[list, str, None] = None,
            sfaira_repo: bool = False,
            cache_path: str = os.path.join('cache', '')
    ):
        self.estimator_embedding = None
        self.estimator_celltype = None
        self.data = None
        self.use_sfaira_repo = sfaira_repo
        self.cache_path = os.path.join(cache_path, '')
        self.adata_ids = AdataIdsSfaira()

        if sfaira_repo:  # check if public sfaira repository should be accessed
            self.model_lookuptable = self._load_lookuptable(settings.sfaira_repo_url)

        if custom_repo:
            if isinstance(custom_repo, str):
                custom_repo = [custom_repo]

            for repo in custom_repo:
                if not os.path.exists(repo):
                    raise OSError(f"provided repo directory does not exist, please create it first: {repo}")

                if not os.path.exists(os.path.join(repo, 'model_lookuptable.csv')):
                    self.write_lookuptable(repo)

                if hasattr(self, 'model_lookuptable'):
                    self.model_lookuptable = self.model_lookuptable.append(
                        self._load_lookuptable(repo)
                    )
                else:
                    self.model_lookuptable = self._load_lookuptable(repo)

        else:
            if not sfaira_repo:
                raise ValueError("please either provide a custom folder/repository with model weights or specify "
                                 "`sfaira_repo=True` to access the public weight repository")

        self.zoo_embedding = ModelZoo(model_lookuptable=self.model_lookuptable, model_class="embedding")
        self.zoo_celltype = ModelZoo(model_lookuptable=self.model_lookuptable, model_class="celltype")

    def _load_lookuptable(
            self,
            repo_path: str
    ):
        """
        checks if there is a csv file that lists the model_id and path of models in the directory
        returns model_lookuptable that connects model_id with the link to the model

        :param repo_path:
        :return: model_lookuptable
        """
        model_lookuptable = pd.read_csv(os.path.join(repo_path, 'model_lookuptable.csv'), header=0, index_col=0)

        # check for any duplicated model_ids
        if hasattr(self, 'model_lookuptable'):
            old_ids = self.model_lookuptable['model_id'].values
            new_ids = model_lookuptable['model_id'].values
            if any(ix in old_ids for ix in new_ids):
                ixs = new_ids[[ix in old_ids for ix in new_ids]]
                raise ValueError('Model ids ' + ','.join(ixs) + ' already exist in a loaded lookuptable.'
                                                                'Please remove any duplicate model_ids')
        return model_lookuptable

    def write_lookuptable(
            self,
            repo_path: str
    ):
        """
        :param repo_path:
        :return:
        """
        import hashlib

        file_names = []
        model_paths = []
        file_paths = []
        md5 = []
        for subdir, dirs, files in os.walk(repo_path):
            for file in files:
                if os.path.isfile(os.path.join(subdir, file)) and (
                        file.endswith('.h5') or file.endswith('.data-00000-of-00001')) and (
                        file.startswith('embedding_') or file.startswith('celltype_')):
                    model_paths.append(os.path.join(subdir, ""))
                    file_paths.append(os.path.join(subdir, file))
                    file_names.append(file)
                    with open(os.path.join(subdir, file), 'rb') as f:
                        md5.append(hashlib.md5(f.read()).hexdigest())
        ids = ['_'.join(i.split('_')[0:3]) for i in file_names]
        ids_cleaned = [i.replace('.h5', '').replace('.data-00000-of-00001', '') for i in ids]  # remove file extensions

        if ids:
            pd.DataFrame(
                list(zip(ids_cleaned, model_paths, file_paths, md5)),
                columns=['model_id', 'model_path', 'model_file_path', 'md5']
            )\
                .sort_values('model_id')\
                .reset_index(drop=True)\
                .to_csv(os.path.join(repo_path, 'model_lookuptable.csv'))
        else:
            raise ValueError(f'No model weights found in {repo_path} '
                             'Weights need to have .h5 or .data-00000-of-00001 extension'
                             'to be recognised'
                             )

    def deposit_zenodo(
            self,
            zenodo_access_token: str,
            title: str,
            authors: list,
            description: str,
            metadata: dict = {},
            update_existing_deposition: Union[None, str] = None,
            publish: bool = False,
            sandbox: bool = False,
            deposit_topologies: bool = True
    ):
        """
        Deposit all models in model lookup table on Zenodo. If publish is set to false, files will be uploaded to a
        deposition draft, which can be further edited (additional metadata, files etc.). Returns the DOI link if
        publish=True or a link to the deposition draft if publish=False.

        :param zenodo_access_token: Your personal Zenodo API access token. Create one here: https://zenodo.org/account/settings/applications/tokens/new/
        :param title: Title of the Zenodo deposition
        :param authors: List of dicts, where each dict defines one author (dict keys:
         name: Name of creator in the format "Family name, Given names",
         affiliation: Affiliation of creator (optional), orcid: ORCID identifier of creator (optional),
         gnd: GND identifier of creator (optional)
        :param description: Description of the Zenodo deposition.
        :param metadata: Dictionary with further metadata attributes of the deposit.
         See the Zenodo API refenrece for accepted keys: https://developers.zenodo.org/#representation
        :param update_existing_deposition: If None, a new deposition will be created.
        If an existing deposition ID is provided as a sting, than this deposition will be updated with a new version.
        :param publish: Set this to True to directly publish the weights on Zenodo.
         When set to False a draft will be created, which can be edited in the browser before publishing.
        :param sandbox: If True, use the Zenodo testing platform at https://sandbox.zenodo.org for your deposition.
         We recommend testing your upload with sandbox first as depositions cannot be deleted from the main Zenodo platform once created.
         :param deposit_topologies: If true, an associated topology file for every weights file will be uploaded to zenodo.
         The naming format for the topology files is <model_id>_topology.pickle
        """

        import requests
        import json
        headers = {"Content-Type": "application/json"}
        params = {'access_token': zenodo_access_token}
        sandbox = 'sandbox.' if sandbox else ''

        # Verify access token
        r = requests.get(f'https://{sandbox}zenodo.org/api/deposit/depositions', params=params)
        if r.status_code != 200:
            raise ValueError(
                "Your Zenodo access token was not accepted by the API. Please provide a valid access token.")

        if update_existing_deposition is None:
            # Create empty deposition
            r = requests.post(f'https://{sandbox}zenodo.org/api/deposit/depositions',
                              params=params,
                              json={},
                              headers=headers)
            # Obtain bucket URL and deposition ID
            bucket_url = r.json()["links"]["bucket"]
            deposition_id = r.json()['id']
        else:
            update_existing_deposition = str(update_existing_deposition) if isinstance(update_existing_deposition, int)\
                else update_existing_deposition
            # Create a new version of the existing deposition
            r = requests.post(
                f'https://{sandbox}zenodo.org/api/deposit/depositions/{update_existing_deposition}/actions/newversion',
                params=params)
            try:
                deposition_id = r.json()["links"]["latest_draft"].split("/")[-1]
            except json.decoder.JSONDecodeError:
                time.sleep(10)
                r = requests.post(
                    f'https://{sandbox}zenodo.org/api/deposit/depositions/{update_existing_deposition}/actions/newversion',
                    params=params)
                deposition_id = r.json()["links"]["latest_draft"].split("/")[-1]
            if r.status_code != 201:
                raise ValueError(
                    f"A new version of deposition {update_existing_deposition} could not be created, "
                    f"please make sure your API key is associated with the account that owns this deposition.")
            r = requests.get(f'https://{sandbox}zenodo.org/api/deposit/depositions/{deposition_id}', params=params)
            bucket_url = r.json()["links"]["bucket"]

            # Delete all existing files from new version
            r_files = requests.get(f'https://{sandbox}zenodo.org/api/deposit/depositions/{deposition_id}/files',
                                   params=params)
            while len(r_files.json()) > 0:
                for file_dict in r_files.json():
                    requests.delete(
                        f'https://{sandbox}zenodo.org/api/deposit/depositions/{deposition_id}/files/{file_dict["id"]}',
                        params=params)
                r_files = requests.get(f'https://{sandbox}zenodo.org/api/deposit/depositions/{deposition_id}/files',
                                       params=params)
                while isinstance(r_files.json(), dict):
                    print("Pausing due to Zenodo API rate limitng")
                    time.sleep(10)
                    r_files = requests.get(f'https://{sandbox}zenodo.org/api/deposit/depositions/{deposition_id}/files',
                                           params=params)

        # Loop over files in model lookup table and upload them one by one
        for i, weight_path in enumerate(self.model_lookuptable['model_file_path']):
            basepath, filename_weights = os.path.split(weight_path)
            with open(weight_path, "rb") as fp:
                r = requests.put(
                    f"{bucket_url}/{filename_weights}",
                    data=fp,
                    params=params,
                )
            while r.status_code != 200:
                print(f"Upload of {weight_path} was not successful (status code {r.status_code}), retrying")
                time.sleep(10)
                with open(weight_path, "rb") as fp:
                    r = requests.put(
                        f"{bucket_url}/{filename_weights}",
                        data=fp,
                        params=params,
                    )
            # Verify checksum after upload
            if r.json()['checksum'][4:] != self.model_lookuptable['md5'][i]:
                warnings.warn(f"The md5 checksum in your model_lookuptable for {self.model_lookuptable['model_id'][i]} "
                              f"does not match the md5 checksum of the uploaded file.")
            if deposit_topologies:  # Deposit associated topology file
                filename_topology = ".".join(filename_weights.split(".")[:-1])
                filename_topology = re.sub(r"_weights$", "", filename_topology)
                filename_topology += "_topology.pickle"
                topology_path = os.path.join(basepath, filename_topology)
                assert os.path.isfile(topology_path), f"topology file {topology_path} not found. " \
                                                      f"consider deactivating the deposition of topology files."
                with open(topology_path, "rb") as fp:
                    r = requests.put(
                        f"{bucket_url}/{filename_topology}",
                        data=fp,
                        params=params,
                    )

        # Add model lookup table to zenodo
        df = self.model_lookuptable.copy()
        df['model_path'] = f"https://{sandbox}zenodo.org/record/{deposition_id}/files/"
        df['model_file_path'] = [f"https://{sandbox}zenodo.org/record/{deposition_id}/files/{os.path.basename(f)}" for f
                                 in self.model_lookuptable['model_file_path']]
        r = requests.put(
            f"{bucket_url}/model_lookuptable.csv",
            data=df.to_csv(),
            params=params,
        )

        # Add metadata
        meta_core = {
            'title': title,
            'creators': authors,
            'description': description,
            'license': 'cc-by-4.0',
            'upload_type': 'dataset',
            'access_right': 'open'
        }
        meta = {**meta_core, **metadata}
        r = requests.put(f'https://{sandbox}zenodo.org/api/deposit/depositions/{deposition_id}',
                         params=params,
                         data=json.dumps({
                             'metadata': meta
                         }),
                         headers=headers)

        if not publish:
            print(f"Zenodo deposition draft has been created: {r.json()['links']['latest_draft_html']}")
            return r.json()['links']['latest_draft_html']
        else:
            # Publish the deposition
            r = requests.post(f'https://{sandbox}zenodo.org/api/deposit/depositions/{deposition_id}/actions/publish',
                              params=params)
            if r.status_code == 202:
                if sandbox:
                    print(f"Weights referenced in model_lookuptable have been sucessfully published on Zenodo: "
                          f"{r.json()['links']['latest_html']}")
                    return r.json()['links']['latest_html']
                else:
                    print(f"Weights referenced in model_lookuptable have been sucessfully published on Zenodo: "
                          f"{r.json()['links']['conceptdoi']}")
                    return r.json()['links']['conceptdoi']
            else:
                try:
                    m = r.json()['message']
                except KeyError:
                    m = f"Submission failed with html status code {r.status_code}"
                raise ValueError(m)

    def load_data(
            self,
            data: anndata.AnnData,
            gene_symbol_col: Union[str, None] = None,
            gene_ens_col: Union[str, None] = None,
            obs_key_celltypes: Union[str, None] = None,
    ):
        """
        Loads the provided AnnData object into sfaira.

        If genes in the provided AnnData object are annotated as gene symbols,
         please provide the name of the corresponding var column (or 'index') through the gene_symbol_col argument.
        If genes in the provided AnnData object are annotated as ensembl ids,
         please provide the name of the corresponding var column (or 'index') through the gene_ens_col argument.
        You need to provide at least one of the two.
        :param data: AnnData object to load
        :param gene_symbol_col: Var column name (or 'index') which contains gene symbols
        :param gene_ens_col: ar column name (or 'index') which contains ensembl ids
        :param obs_key_celltypes: .obs column name which contains cell type labels.
        """
        if self.zoo_embedding.model_organism is not None and self.zoo_celltype.model_organism is not None:
            assert self.zoo_embedding.model_organism == self.zoo_celltype.model_organism, \
                "Model ids set for embedding and celltype model need to correspond to the same organism"
            assert self.zoo_embedding.model_organ == self.zoo_celltype.model_organ, \
                "Model ids set for embedding and celltype model need to correspond to the same organ"
        if self.zoo_embedding.model_organism is not None:
            organism = self.zoo_embedding.model_organism
            organ = self.zoo_embedding.model_organ
        elif self.zoo_celltype.model_organism is not None:
            organism = self.zoo_celltype.model_organism
            organ = self.zoo_celltype.model_organ
        else:
            raise ValueError("Please first set which model_id to use via the model zoo before loading the data")

        if gene_ens_col is None and gene_symbol_col is None:
            raise ValueError("Please provide either the gene_ens_col or the gene_symbol_col argument.")

        # handle organ names with stripped spaces
        if organ not in OCS.organ.node_names:
            organ_dict = {i.replace(" ", ""): i for i in OCS.organ.node_names}
            assert organ in organ_dict, f"Organ {organ} is not a valid nodename in the UBERON organ ontology"
            organ = {i.replace(" ", ""): i for i in OCS.organ.node_names}[organ]

        self.data = DatasetInteractive(
            data=data,
            feature_symbol_col=gene_symbol_col,
            feature_id_col=gene_ens_col,
        )
        self.data.organism = organism
        self.data.organ = organ
        self.data.cell_type_obs_key = obs_key_celltypes
        # Align to correct featurespace
        self.data.streamline_features(
            match_to_release=self.zoo_embedding.topology_container.gc.release,
            subset_genes_to_type=list(set(self.zoo_embedding.topology_container.gc.biotype))
        )
        # Transfer required metadata from the Dataset instance to the adata object
        self.data.streamline_metadata(
            clean_obs=False,
            clean_var=True,
            clean_uns=False,
            clean_obs_names=False,
        )

    def _load_topology_dict(self, model_weights_file) -> dict:
        topology_filepath = ".".join(model_weights_file.split(".")[:-1])
        topology_filepath = re.sub(r"_weights$", "", topology_filepath)
        topology_filepath += "_topology.pickle"
        if topology_filepath.startswith('http'):
            # Download into cache if file is on a remote server.
            if not os.path.exists(self.cache_path):
                os.makedirs(self.cache_path)
            import urllib.request
            from urllib.error import HTTPError
            try:
                urllib.request.urlretrieve(
                    topology_filepath,
                    os.path.join(self.cache_path, os.path.basename(topology_filepath))
                )
                topology_filepath = os.path.join(self.cache_path, os.path.basename(topology_filepath))
            except HTTPError:
                raise FileNotFoundError(f"cannot find remote topology file {topology_filepath}")
        with open(topology_filepath, "rb") as f:
            topology = pickle.load(f)
        return topology

    def load_model_embedding(self):
        """
        Initialise embedding model and load parameters from public parameter repository.

        Loads model defined in self.model_id.

        :return: Model ID loaded.
        """
        assert self.zoo_embedding.model_id is not None, "choose embedding model first"
        if self.zoo_celltype.topology_container.gc.release is not None:
            assert self.zoo_embedding.topology_container.gc.release == \
                self.zoo_celltype.topology_container.gc.release, \
                "genome assemblies defined in the topology " \
                "containers if the embedding and the celltype " \
                "prediction model are not equivalent " \
                f"({self.zoo_embedding.topology_container.gc.release} " \
                f"and {self.zoo_celltype.topology_container.gc.release} " \
                f"respectively, aborting.)"
        model_weights_file = self.model_lookuptable["model_file_path"].loc[self.model_lookuptable["model_id"] ==
                                                                           self.zoo_embedding.model_id].iloc[0]
        md5 = self.model_lookuptable["md5"].loc[self.model_lookuptable["model_id"] ==
                                                self.zoo_embedding.model_id].iloc[0]
        tc = TopologyContainer(
            topology=self._load_topology_dict(model_weights_file=model_weights_file),
            topology_id=self.zoo_embedding.topology_container.topology_id
        )
        self.estimator_embedding = EstimatorKerasEmbedding(
            data=self.data.adata,
            model_dir=model_weights_file,
            model_id=self.zoo_embedding.model_id,
            model_topology=tc,
            weights_md5=md5,
            cache_path=self.cache_path,
            adata_ids=self.adata_ids
        )
        self.estimator_embedding.init_model()
        self.estimator_embedding.load_pretrained_weights()

    def load_model_celltype(self):
        """
        Initialise cell type model and load parameters from public parameter repository.

        Loads model defined in self.model_id.

        :return: Model ID loaded.
        """
        assert self.zoo_celltype.model_id is not None, "choose cell type model first"
        if self.zoo_embedding.topology_container.gc.release is not None:
            assert self.zoo_embedding.topology_container.gc.release == \
                self.zoo_celltype.topology_container.gc.release, \
                "genome assemblies defined in the topology " \
                "containers if the embedding and the celltype " \
                "prediction model are not equivalent " \
                f"({self.zoo_embedding.topology_container.gc.release} " \
                f"and {self.zoo_celltype.topology_container.gc.release} " \
                f"respectively, aborting.)"
        model_weights_file = self.model_lookuptable["model_file_path"].loc[self.model_lookuptable["model_id"] ==
                                                                           self.zoo_celltype.model_id].iloc[0]
        md5 = self.model_lookuptable["md5"].loc[self.model_lookuptable["model_id"] ==
                                                self.zoo_celltype.model_id].iloc[0]
        tc = TopologyContainer(
            topology=self._load_topology_dict(model_weights_file=model_weights_file),
            topology_id=self.zoo_celltype.topology_container.topology_id
        )
        self.estimator_celltype = EstimatorKerasCelltype(
            data=self.data.adata,
            model_dir=model_weights_file,
            model_id=self.zoo_celltype.model_id,
            model_topology=tc,
            weights_md5=md5,
            cache_path=self.cache_path,
            remove_unlabeled_cells=False,
            adata_ids=self.adata_ids
        )
        self.estimator_celltype.init_model()
        self.estimator_celltype.load_pretrained_weights()

    def _adata_write_celltype(
            self,
            labels: List[str],
            key: str
    ):
        """
        Writes a list of cell type labels into the column of adata.obs indicated
        :return:
        """
        key_id = key + "_id"
        self.data.adata.obs[key] = [self.estimator_celltype.ontology_names[i] for i in np.argmax(labels, axis=1)]
        self.data.adata.obs[key] = self.data.adata.obs[key].astype('category')
        self.data.adata.obs[key_id] = [self.estimator_celltype.ontology_ids[i] for i in np.argmax(labels, axis=1)]
        self.data.adata.obs[key_id] = self.data.adata.obs[key_id].astype('category')

    def _adata_write_embedding(
            self,
            embedding: List[str],
            key: str
    ):
        """
        Writes the embedding matrix into adata.obsm with the key indicated.
        :return:
        """
        self.data.adata.obsm[key] = embedding

    def _adata_write_denoised_data(
            self,
            denoised_data: np.ndarray,
            key: str
    ):
        """
        Writes the denoised expression matrix into adata.obsm with the key indicated.
        :return:
        """
        self.data.adata.layers[key] = denoised_data

    def predict_celltypes(self):
        """
        Run local cell type prediction model and add predictions to adata.obs.

        :return:
        """
        if self.zoo_celltype is not None:
            self._adata_write_celltype(
                labels=self.estimator_celltype.predict(),
                key="celltypes_sfaira"
            )
        else:
            raise ValueError("celltype zoo has to be set before local model can be run.")

    def predict_embedding(self):
        """
        Run local embedding prediction model and add embedding to adata.obsm.

        :return:
        """
        if self.zoo_embedding is not None:
            self._adata_write_embedding(
                embedding=self.estimator_embedding.predict_embedding(),
                key="X_sfaira"
            )
        else:
            raise ValueError("embedding zoo has to be set before local model can be run.")

    def predict_all(self):
        """
        Run local cell type prediction and embedding models and add results of both to adata.

        :return:
        """
        self.predict_embedding()
        self.predict_celltypes()

    def compute_denoised_expression(self):
        """
        Run local embedding prediction model and add denoised expression to new adata layer.

        :return:
        """
        if self.zoo_embedding is not None:
            self._adata_write_denoised_data(
                denoised_data=self.estimator_embedding.predict(),
                key="denoised_sfaira"
            )
        else:
            raise ValueError("embedding zoo has to be set before local model can be run.")

    def celltype_summary(self):
        """
        Return type with frequencies of predicted cell types.

        :return:
        """
        assert "celltypes_sfaira" in self.data.adata.obs.keys(), \
            "Column celltypes_sfaira not found in the data. Please run UserInterface.predict_celltypes() first."
        return self.data.adata.obs['celltypes_sfaira'].value_counts()



from typing import Iterable, List, Union

from sfaira.versions.genomes import GenomeContainer


def translate_symbols_to_id(x: Union[str, Iterable[str]], release: str, organism: str) -> Union[str, List[str]]:
    """
    Translate gene symbols to ENSEMBL IDs.

    Input captitalization is ignored but the output capitalisation matches the ENSEMBL .gtf files.

    Are you not sure which assembly to use?

    - You could use the newest one for example, check the ENSEMBL site regularly for updates:
        http://ftp.ensembl.org/pub/
    - You could use one used by a specific aligner, the assemblies used by 10x cellranger are described here
        for example: https://support.10xgenomics.com/single-cell-gene-expression/software/release-notes/build

    :param x: Symbol(s) to translate.
    :param release: The name of the release of genome assembly, e.g. "102" for assembly "Homo_sapiens.GRCh38.102".
    :param organism: The name of the organism of genome assembly, e.g. "Homo sapiens" for assembly
        "Homo_sapiens.GRCh38.102".
    :return: ENSEMBL IDs
    """
    return GenomeContainer(release=release, organism=organism).translate_symbols_to_id(x=x)


def translate_id_to_symbols(x: Union[str, Iterable[str]], release: str, organism: str) -> Union[str, List[str]]:
    """
    Translate ENSEMBL IDs to gene symbols.

    Input captitalization is ignored but the output capitalisation matches the ENSEMBL .gtf files.

    Are you not sure which assembly to use?

    - You could use the newest one for example, check the ENSEMBL site regularly for updates:
        http://ftp.ensembl.org/pub/
    - You could use one used by a specific aligner, the assemblies used by 10x cellranger are described here
        for example: https://support.10xgenomics.com/single-cell-gene-expression/software/release-notes/build

    :param x: ENSEMBL ID(s) to translate.
    :param release: The name of the release of genome assembly, e.g. "102" for assembly "Homo_sapiens.GRCh38.102".
    :param organism: The name of the organism of genome assembly, e.g. "Homo sapiens" for assembly
        "Homo_sapiens.GRCh38.102".
    :return: Gene symbols.
    """
    return GenomeContainer(release=release, organism=organism).translate_id_to_symbols(x=x)



import logging

import anndata
import numpy as np
from rich import print
import scipy.sparse

log = logging.getLogger(__name__)


class H5adValidator:

    def __init__(self, test_h5ad, schema=None):
        self.fn_h5ad: str = test_h5ad
        if schema not in ["cellxgene"]:
            raise ValueError(f"Did not recognize schema {schema}")
        self.schema = schema
        self._adata = None

    @property
    def adata(self):
        if self._adata is None:
            self._adata = anndata.read_h5ad(filename=self.fn_h5ad)
        return self._adata

    def validate(self):
        self._test_schema()
        self._test_numeric_data()
        print(f'[bold orange]Sfaira butler: "The object {self.fn_h5ad} passed the schema {self.schema}."')

    def _test_schema(self) -> None:
        """Verify that object elements match schema definitions."""
        if self.schema == "cellxgene":
            from cellxgene_schema import validate
            validate.validate(h5ad_path=self.fn_h5ad)
        else:
            raise ValueError(f"schema {self.schema} not supported")

    def _test_numeric_data(self) -> None:
        """Verify that numeric matrices match schema definitions."""
        # if isinstance(self.adata.X, scipy.sparse.spmatrix):
        #     x = np.unique(np.asarray(self.adata.X.todense()))
        # else:
        #     x = np.unique(np.asarray(self.adata.X))
        # deviation_from_integer = np.minimum(x % 1, 1. - x % 1)
        # assert np.max(deviation_from_integer) < 1e-6
        pass



import os
import pathlib

from sfaira.data import DatasetSuperGroup
from sfaira.data.dataloaders.databases import DatasetSuperGroupDatabases
from sfaira.unit_tests.data_for_tests.databases.consts import CELLXGENE_COLLECTION_ID

from sfaira.unit_tests.directories import DIR_DATA_DATABASES_CACHE


def prepare_dsg_database(database: str, download: bool = True) -> DatasetSuperGroup:
    """
    Prepares data set super group of data base returns instance.

    :param database: Database to make available.
    :param download: Whether to make sure that raw files are downloaded.
    """
    if not os.path.exists(DIR_DATA_DATABASES_CACHE):
        pathlib.Path(DIR_DATA_DATABASES_CACHE).mkdir(parents=True, exist_ok=True)
    if database == "cellxgene":
        # Note: if unit tests related to cellxgene throw error because of caching, empty cache.
        # Unit tests are slow if cache_metadata is False.
        dsg = DatasetSuperGroupDatabases(data_path=DIR_DATA_DATABASES_CACHE, cache_metadata=True)
        # Only retain pre-defined target collections to avoid bulk downloads during unit tests.
        dsg.subset(key="collection_id", values=CELLXGENE_COLLECTION_ID)
    else:
        assert False, database
    if download:
        dsg.download()
    return dsg



import logging
import os
import sys

import click
import rich
import rich.logging
from rich import traceback
from rich import print

from sfaira.commands.annotate_dataloader import DataloaderAnnotater
from sfaira.commands.cache_control import CacheControl
from sfaira.commands.export_h5ad import H5adExport
from sfaira.commands.submit_pullrequest import PullRequestHandler
from sfaira.commands.test_dataloader import DataloaderTester
from sfaira.commands.utils import doi_lint
from sfaira.commands.validate_dataloader import DataloaderValidator
from sfaira.commands.validate_h5ad import H5adValidator

from sfaira import __version__
from sfaira.commands.create_dataloader import DataloaderCreator
from sfaira.commands.upgrade import UpgradeCommand

WD = os.path.dirname(__file__)
log = logging.getLogger()

DEFAULT_DATA_PATH = "./"
PACKAGE_LOADER_PATH = "sfaira/data/dataloaders/loaders/"


def check_paths(pathlist):
    for p in pathlist:
        if p is None or not os.path.isdir(p):
            print(f"Error: Path {p} does not exist.")
            sys.exit()


def set_paths(loader=None, data=None, cache=None):
    env_loader_path = os.getenv('SFAIRA_LOADER_PATH')
    env_data_path = os.getenv('SFAIRA_DATA_PATH')
    env_cache_path = os.getenv('SFAIRA_CACHE_PATH')
    if env_loader_path:
        if not os.getenv('SFAIRA_DOCKER'):  # Skip print if running in sfaira docker where env variables are always set
            print('[bold blue]SFAIRA_LOADER_PATH environment variable detected. Ignoring --path-loader if supplied.')
        loader = env_loader_path
    if env_data_path:
        if not os.getenv('SFAIRA_DOCKER'):  # Skip print if running in sfaira docker where env variables are always set
            print('[bold blue]SFAIRA_DATA_PATH environment variable detected. Ignoring --path-data if supplied.')
        data = env_data_path
    if env_cache_path:
        if not os.getenv('SFAIRA_DOCKER'):  # Skip print if running in sfaira docker where env variables are always set
            print('[bold blue]SFAIRA_CACHE_PATH environment variable detected. Ignoring --path-cache if supplied.')
        cache = env_cache_path
    check_paths([loader, data])
    return loader, data, cache


def main():
    traceback.install(width=200, word_wrap=True)
    print(r"""[bold blue]
███████ ███████  █████  ██ ██████   █████ 
██      ██      ██   ██ ██ ██   ██ ██   ██ 
███████ █████   ███████ ██ ██████  ███████ 
     ██ ██      ██   ██ ██ ██   ██ ██   ██ 
███████ ██      ██   ██ ██ ██   ██ ██   ██ 
                                           
        """)

    print('[bold blue]Run [green]sfaira --help [blue]for an overview of all commands\n')

    # Is the latest sfaira version installed? Upgrade if not!
    if not UpgradeCommand.check_sfaira_latest():
        print('[bold blue]Run [green]sfaira upgrade [blue]to get the latest version.')
    sfaira_cli()


@click.group()
@click.version_option(__version__, message=click.style(f'sfaira Version: {__version__}', fg='blue'))
@click.option('-v', '--verbose', is_flag=True, default=False, help='Enable verbose output (print debug statements).')
@click.option("-l", "--log-file", help="Save a verbose log to a file.")
@click.pass_context
def sfaira_cli(ctx, verbose, log_file):
    """
    Create and manage sfaira dataloaders.
    """
    # Set the base logger to output DEBUG
    log.setLevel(logging.DEBUG)

    # Set up logs to the console
    log.addHandler(
        rich.logging.RichHandler(
            level=logging.DEBUG if verbose else logging.INFO,
            console=rich.console.Console(file=sys.stderr),
            show_time=True,
            markup=True,
        )
    )

    # Set up logs to a file if we asked for one
    if log_file:
        log_fh = logging.FileHandler(log_file, encoding="utf-8")
        log_fh.setLevel(logging.DEBUG)
        log_fh.setFormatter(logging.Formatter("[%(asctime)s] %(name)-20s [%(levelname)-7s]  %(message)s"))
        log.addHandler(log_fh)


@sfaira_cli.command()
def cache_clear() -> None:
    """Clears sfaira cache, including ontology and genome cache."""
    ctrl = CacheControl()
    ctrl.clear()


@sfaira_cli.command()
def cache_reload() -> None:
    """Downloads new ontology versions into cache."""
    ctrl = CacheControl()
    ctrl.reload()


@sfaira_cli.command()
@click.option('--path-data',
              default=DEFAULT_DATA_PATH,
              type=click.Path(exists=False),
              help='Absolute path of the desired location of the raw data directory.')
@click.option('--path-loader',
              default=PACKAGE_LOADER_PATH,
              type=click.Path(exists=False),
              help='Relative path from the current directory to the desired location of the data loader.')
def create_dataloader(path_data, path_loader) -> None:
    """
    Interactively create a new sfaira dataloader.
    """
    path_loader, path_data, _ = set_paths(loader=path_loader, data=path_data)
    dataloader_creator = DataloaderCreator(path_loader)
    dataloader_creator.create_dataloader(path_data)


@sfaira_cli.command()
@click.option('--doi', type=str, required=True, help="The doi of the paper that the data loader refers to.")
@click.option('--path-data',
              default=DEFAULT_DATA_PATH,
              type=click.Path(exists=False),
              help='Absolute path of the location of the raw data directory.')
@click.option('--path-loader',
              default=PACKAGE_LOADER_PATH,
              type=click.Path(exists=False),
              help='Relative path from the current directory to the location of the data loader.')
@click.option('--schema', type=str, default="sfaira", help="The curation schema to check meta data availability for.")
def annotate_dataloader(doi, path_data, path_loader, schema) -> None:
    """
    Annotates a dataloader.
    """
    path_loader, path_data, _ = set_paths(loader=path_loader, data=path_data)
    if doi_lint(doi):
        dataloader_validator = DataloaderValidator(path_loader, doi, schema=schema)
        dataloader_validator.validate()
        dataloader_annotater = DataloaderAnnotater()
        dataloader_annotater.annotate(path_loader, path_data, dataloader_validator.doi)
    else:
        print('[bold red]The supplied DOI is malformed!')  # noqa: W605


@sfaira_cli.command()
@click.option('--doi', type=str, required=True, help="The doi of the paper that the data loader refers to.")
@click.option('--path-loader',
              default=PACKAGE_LOADER_PATH,
              type=click.Path(exists=False),
              help='Relative path from the current directory to the desired location of the data loader.')
@click.option('--schema', type=str, default="sfaira", help="The curation schema to check meta data availability for.")
def validate_dataloader(doi, path_loader, schema) -> None:
    """
    Verifies the dataloader against sfaira's requirements.
    """
    path_loader, _, _ = set_paths(loader=path_loader)
    if doi_lint(doi):
        dataloader_validator = DataloaderValidator(path_loader=path_loader, doi=doi, schema=schema)
        dataloader_validator.validate()
    else:
        print('[bold red]The supplied DOI is malformed!')  # noqa: W605


def _full_test(path_loader, path_data, doi, schema, clean_tsvs, in_phase_3):
    if doi is None or doi_lint(doi):
        dataloader_validator = DataloaderValidator(path_loader=path_loader, doi=doi, schema=schema)
        dataloader_validator.validate()
        dataloader_tester = DataloaderTester(path_loader, path_data, doi)
        dataloader_tester.test_dataloader(clean_tsvs=clean_tsvs, in_phase_3=in_phase_3)
    else:
        print('[bold red]The supplied DOI is malformed!')  # noqa: W605


@sfaira_cli.command()
@click.option('--doi', type=str, required=True, help="The doi of the paper that the data loader refers to.")
@click.option('--path-data',
              default=DEFAULT_DATA_PATH,
              type=click.Path(exists=False),
              help='Absolute path of the location of the raw data directory.')
@click.option('--path-loader',
              default=PACKAGE_LOADER_PATH,
              type=click.Path(exists=False),
              help='Relative path from the current directory to the location of the data loader.')
@click.option('--schema', type=str, default="sfaira", help="The curation schema to check meta data availability for.")
def test_dataloader(doi, path_data, path_loader, schema) -> None:
    """
    Runs a full data loader test.
    """
    path_loader, path_data, _ = set_paths(loader=path_loader, data=path_data)
    _full_test(path_loader=path_loader, path_data=path_data, doi=doi, schema=schema, clean_tsvs=False, in_phase_3=False)


@sfaira_cli.command()
@click.option('--doi', type=str, required=True, help="The doi of the paper that the data loader refers to.")
@click.option('--path-data',
              default=DEFAULT_DATA_PATH,
              type=click.Path(exists=False),
              help='Absolute path of the location of the raw data directory.')
@click.option('--path-loader',
              default=PACKAGE_LOADER_PATH,
              type=click.Path(exists=False),
              help='Relative path from the current directory to the location of the data loader.')
@click.option('--schema', type=str, default="sfaira", help="The curation schema to check meta data availability for.")
def finalize_dataloader(doi, path_data, path_loader, schema) -> None:
    """
    Formats .tsvs and runs a full data loader test.
    """
    path_loader, path_data, _ = set_paths(loader=path_loader, data=path_data)
    _full_test(path_loader=path_loader, path_data=path_data, doi=doi, schema=schema, clean_tsvs=True, in_phase_3=True)


@sfaira_cli.command()
def publish_dataloader() -> None:
    """
    Interactively create a GitHub pull request for a newly created data loader.
    This only works when called in the sfaira CLI docker container.
    Runs a full data loader test before starting the pull request.
    """
    path_loader, _, _ = set_paths()
    pullrequest_handler = PullRequestHandler(path_loader)
    pullrequest_handler.submit_pr()


@sfaira_cli.command()
@click.option('--doi', type=str, required=True, help="The doi of the paper that the data loader refers to.")
@click.option('--schema', type=str, default=None, help="Schema to streamline to, e.g. 'cellxgene'")
@click.option('--path-out',
              type=click.Path(exists=True),
              help='Absolute path of the location of the streamlined output h5ads.')
@click.option('--path-data',
              default=DEFAULT_DATA_PATH,
              type=click.Path(exists=True),
              help='Absolute path of the location of the raw data directory.')
@click.option('--path-loader',
              default=PACKAGE_LOADER_PATH,
              type=click.Path(exists=False),
              help='Relative path from the current directory to the location of the data loader.')
@click.option('--path-cache',
              type=click.Path(exists=True),
              default=None,
              help='The optional absolute path to cached data library maintained by sfaira. Using such a cache speeds '
                   'up loading in sequential runs but is not necessary.')
def export_h5ad(doi, schema, path_out, path_data, path_loader, path_cache) -> None:
    """Creates a collection of streamlined h5ad object for a given DOI."""
    path_loader, path_data, _ = set_paths(loader=path_loader, data=path_data, cache=path_cache)
    h5ad_export = H5adExport(doi=doi, path_cache=path_cache, path_data=path_data, path_loader=path_loader,
                             path_out=path_out, schema=schema)
    h5ad_export.write()


@sfaira_cli.command()
@click.option('--h5ad', type=click.Path(exists=True))
@click.option('--schema', type=str, default=None)
def validate_h5ad(h5ad, schema) -> None:
    """Runs a component test on a streamlined h5ad object.

    h5ad is the absolute path of the .h5ad file to test.
    schema is the schema type ("cellxgene",) to test.
    """
    h5ad_validator = H5adValidator(h5ad, schema)
    h5ad_validator.validate()


if __name__ == "__main__":
    traceback.install()
    sys.exit(main())  # pragma: no cover



from sfaira.versions.topologies.musmusculus.embedding.ae import AE_TOPOLOGIES
from sfaira.versions.topologies.musmusculus.embedding.linear import LINEAR_TOPOLOGIES
from sfaira.versions.topologies.musmusculus.embedding.nmf import NMF_TOPOLOGIES
from sfaira.versions.topologies.musmusculus.embedding.vae import VAE_TOPOLOGIES
from sfaira.versions.topologies.musmusculus.embedding.vaeiaf import VAEIAF_TOPOLOGIES
from sfaira.versions.topologies.musmusculus.embedding.vaevamp import VAEVAMP_TOPOLOGIES



"""
Functionalities to interact with feature sets defined in an assembly or interactively by user.
"""
import abc
import ftplib
import gzip
import numpy as np
import os
from typing import Iterable, List, Union
import pandas
import pathlib
import urllib.error
import urllib.request

from sfaira import settings

KEY_SYMBOL = "gene_name"
KEY_ID = "gene_id"
KEY_TYPE = "gene_biotype"
VALUE_GTF_GENE = "gene"
KEY_GTF_REGION_TYPE = 2


class GtfInterface:

    release: str
    organism: str

    def __init__(self, release: str, organism: str):
        self.release = release
        self.organism = organism

    @property
    def cache_dir(self):
        """
        The cache dir is in a cache directory in the homedirectory of the user by default and can be user modified.
        """
        cache_dir_path = os.path.join(settings.cachedir_genomes, self.ensembl_organism)
        cache_dir_path = pathlib.Path(cache_dir_path)
        cache_dir_path.mkdir(parents=True, exist_ok=True)
        return cache_dir_path

    @property
    def cache_fn(self):
        return os.path.join(self.cache_dir, self.assembly + ".csv")

    @property
    def assembly(self) -> str:
        """
        Get full assembly name of genome represented in this class.
        """
        def file_filter(fns) -> List[str]:
            # Filter assembly files starting with organism name:
            y = [x for x in fns if x.split(".")[0].lower() == self.ensembl_organism]
            # Filter target release:
            y = [x for x in y if x.split(".")[-2] == self.release]
            return y

        # Check first in local files and then query ensembl if no match is found:
        target_file = file_filter(fns=os.listdir(self.cache_dir))
        if len(target_file) == 0:
            # Get variable middle string of assembly name by looking files up on ftp server:
            ftp = ftplib.FTP("ftp.ensembl.org")
            ftp.login()
            ftp.cwd(self.url_ensembl_dir)
            data = []
            ftp.dir(data.append)
            ftp.quit()
            target_file = [line.split(' ')[-1] for line in data]
            # Filter assembly files starting with organism name:
            target_file = [x for x in target_file if x.split(".")[0].lower() == self.ensembl_organism]
            # Filter target assembly:
            target_file = [x for x in target_file if x.endswith(f".{self.release}.gtf.gz")]
            # There should only be one file left if filters work correctly:
            assert len(target_file) == 1, target_file
            assembly = target_file[0].split(".gtf.gz")[0]
        else:
            assert len(target_file) == 1, target_file  # There should only be one file left if filters work correctly.
            assembly = target_file[0].split(".csv")[0]
        return assembly

    @property
    def ensembl_organism(self):
        return "_".join([x.lower() for x in self.organism.split(" ")])

    @property
    def url_ensembl_dir(self):
        return f"pub/release-{self.release}/gtf/{self.ensembl_organism}"

    @property
    def url_ensembl_gtf(self):
        return f"ftp://ftp.ensembl.org/{self.url_ensembl_dir}/{self.assembly}.gtf.gz"

    def download_gtf_ensembl(self):
        """
        Download .gtf file from ensembl FTP server and turn into reduced, gene-centric cache .csv.
        """
        temp_file = os.path.join(self.cache_dir, self.assembly + ".gtf.gz")
        try:
            _ = urllib.request.urlretrieve(url=self.url_ensembl_gtf, filename=temp_file)
        except urllib.error.URLError as e:
            raise ValueError(f"Could not download gtf from {self.url_ensembl_gtf} with urllib.error.URLError: {e}, "
                             f"check if assembly name '{self.assembly}' corresponds to an actual assembly.")
        with gzip.open(temp_file) as f:
            tab = pandas.read_csv(f, sep="\t", comment="#", header=None)
        # Delete temporary file .gtf.gz if it still exists (some times already deleted by parallel process in grid
        # search).
        if os.path.exists(temp_file):
            os.remove(temp_file)
        tab = tab.loc[tab[KEY_GTF_REGION_TYPE].values == VALUE_GTF_GENE, :]
        # KEY_SYMBOL column: Uses symbol if available, otherwise replaces with ID.
        conversion_tab = pandas.DataFrame({
            KEY_ID: [
                x.split(KEY_ID)[1].split(";")[0].strip(" ").strip("\"")
                for x in tab.iloc[:, -1].values],
            KEY_SYMBOL: [
                x.split(KEY_SYMBOL)[1].split(";")[0].strip(" ").strip("\"")
                if KEY_SYMBOL in x else
                x.split(KEY_ID)[1].split(";")[0].strip(" ").strip("\"")
                for x in tab.iloc[:, -1].values],
            KEY_TYPE: [
                x.split(KEY_TYPE)[1].split(";")[0].strip(" ").strip("\"")
                for x in tab.iloc[:, -1].values],
        }).sort_values(KEY_ID)
        conversion_tab.to_csv(self.cache_fn)

    @property
    def cache(self) -> pandas.DataFrame:
        if not os.path.exists(self.cache_fn):
            self.download_gtf_ensembl()
        return pandas.read_csv(self.cache_fn)


class GenomeContainerBase:
    """
    Base container class for a gene annotation.
    """

    @abc.abstractmethod
    def organism(self):
        pass

    @abc.abstractmethod
    def set(self, **kwargs):
        pass

    @abc.abstractmethod
    def symbols(self) -> List[str]:
        pass

    @abc.abstractmethod
    def ensembl(self) -> List[str]:
        pass

    @abc.abstractmethod
    def biotype(self) -> List[str]:
        pass

    @abc.abstractmethod
    def n_var(self):
        pass


class GenomeContainer(GenomeContainerBase):
    """
    Container class for a genome annotation for a specific release.

    This class can be used to translate between symbols and ENSEMBL IDs for a specific assembly, to store specific gene
    subsets of an assembly, and to subselect genes by biotypes in an assembly.
    """

    genome_tab: pandas.DataFrame
    organism: str
    release: str

    def __init__(
            self,
            organism: str = None,
            release: str = None,
    ):
        """
        Are you not sure which assembly to use?

            - You could use the newest one for example, check the ENSEMBL site regularly for updates:
                http://ftp.ensembl.org/pub/
            - You could use one used by a specific aligner, the assemblies used by 10x cellranger are described here
                for example: https://support.10xgenomics.com/single-cell-gene-expression/software/release-notes/build

        :param release: The full name of the genome assembly, e.g. Homo_sapiens.GRCh38.102.
        """
        if not isinstance(organism, str):
            raise ValueError(f"supplied organism {organism} was not a string")
        if not isinstance(release, str):
            raise ValueError(f"supplied release {release} was not a string")
        self.organism = organism
        self.release = release
        self.gtfi = GtfInterface(organism=self.organism, release=self.release)
        self.load_genome()

    def load_genome(self):
        self.genome_tab = self.gtfi.cache

    def set(
            self,
            biotype: Union[None, str, List[str]] = None,
            symbols: Union[None, str, List[str]] = None,
            ensg: Union[None, str, List[str]] = None,
    ):
        """
        Subset by gene biotype or to gene list defined by identifiers (symbol or ensemble ID).

        Will subset by multiple factors if more than one parameter is not None.

        :param biotype: Gene biotype(s) of gene(s) to subset genome to. Elements have to appear in genome.
            Separate in string via "," if choosing multiple or supply as list of string.
        :param symbols: Gene symbol(s) of gene(s) to subset genome to. Elements have to appear in genome.
            Separate in string via "," if choosing multiple or supply as list of string.
        :param ensg: Ensemble gene ID(s) of gene(s) to subset genome to. Elements have to appear in genome.
            Separate in string via "," if choosing multiple or supply as list of string.
        """
        subset = np.ones((self.n_var,), "int") == 1
        if biotype is not None:
            if isinstance(biotype, list):
                pass
            elif isinstance(biotype, str):
                biotype = biotype.split(",")
            else:
                raise ValueError(f"Supply biotype as string, see also function annotation. Supplied {biotype}.")
            self.__validate_types(x=biotype)
            subset = np.logical_and(
                subset,
                [x in biotype for x in self.genome_tab[KEY_TYPE].values]
            )
        if symbols is not None:
            if isinstance(symbols, list):
                pass
            elif isinstance(symbols, str):
                symbols = symbols.split(",")
            else:
                raise ValueError(f"Supply symbols as string, see also function annotation. Supplied {symbols}.")
            self.__validate_symbols(x=symbols)
            subset = np.logical_and(
                subset,
                [x in symbols for x in self.genome_tab[KEY_SYMBOL].values]
            )
        if ensg is not None:
            if isinstance(ensg, list):
                pass
            elif isinstance(ensg, str):
                ensg = ensg.split(",")
            else:
                raise ValueError(f"Supply ensg as string, see also function annotation. Supplied {ensg}.")
            self.__validate_ensembl(x=ensg)
            subset = np.logical_and(
                subset,
                [x in ensg for x in self.genome_tab[KEY_ID].values]
            )
        self.genome_tab = self.genome_tab.loc[subset, :].copy()

    @property
    def symbols(self) -> List[str]:
        """
        List of symbols of genes in genome container.
        """
        return self.genome_tab[KEY_SYMBOL].values.tolist()

    @property
    def ensembl(self) -> List[str]:
        """
        List of ENSEMBL IDs of genes in genome container.
        """
        return self.genome_tab[KEY_ID].values.tolist()

    @property
    def biotype(self) -> List[str]:
        """
        List of biotypes of genes in genome container.
        """
        return self.genome_tab[KEY_TYPE].values.tolist()

    def __validate_ensembl(self, x: List[str], enforce_captitalization: bool = True):
        if enforce_captitalization:
            not_found = [y for y in x if y not in self.ensembl]
        else:
            ensembl_upper = [y.upper() for y in self.ensembl]
            not_found = [y for y in x if y.upper() not in ensembl_upper]
        if len(not_found) > 0:
            raise ValueError(f"Could not find ENSEMBL ID: {not_found}")

    def __validate_symbols(self, x: List[str], enforce_captitalization: bool = True):
        if enforce_captitalization:
            not_found = [y for y in x if y not in self.symbols]
        else:
            symbols_upper = [y.upper() for y in self.symbols]
            not_found = [y for y in x if y.upper() not in symbols_upper]
        if len(not_found) > 0:
            raise ValueError(f"Could not find symbol: {not_found}")

    def __validate_types(self, x: List[str]):
        not_found = [y for y in x if y not in self.biotype]
        if len(not_found) > 0:
            raise ValueError(f"Could not find type: {not_found}")

    @property
    def n_var(self) -> int:
        """
        Number of genes in genome container.
        """
        return self.genome_tab.shape[0]

    @property
    def symbol_to_id_dict(self):
        """
        Dictionary-formatted map of gene symbols to ENSEMBL IDs.
        """
        return dict(zip(self.genome_tab[KEY_SYMBOL].values.tolist(), self.genome_tab[KEY_ID].values.tolist()))

    @property
    def id_to_symbols_dict(self):
        """
        Dictionary-formatted map of ENSEMBL IDs to gene symbols.
        """
        return dict(zip(self.genome_tab[KEY_ID].values.tolist(), self.genome_tab[KEY_SYMBOL].values.tolist()))

    def translate_symbols_to_id(self, x: Union[str, Iterable[str]]) -> Union[str, List[str]]:
        """
        Translate gene symbols to ENSEMBL IDs.

        :param x: Symbol(s) to translate.
        :return: ENSEMBL IDs
        """
        if isinstance(x, str):
            x = [x]
        self.__validate_symbols(x=x, enforce_captitalization=False)
        map_dict = dict([(k.upper(), v) for k, v in self.symbol_to_id_dict.items()])
        y = [map_dict[xx.upper()] for xx in x]
        if len(y) == 1:
            y = y[0]
        return y

    def translate_id_to_symbols(self, x: Union[str, Iterable[str]]) -> Union[str, List[str]]:
        """
        Translate ENSEMBL IDs to gene symbols.

        :param x: ENSEMBL ID(s) to translate.
        :return: Gene symbols.
        """
        if isinstance(x, str):
            x = [x]
        self.__validate_ensembl(x=x, enforce_captitalization=False)
        map_dict = dict([(k.upper(), v) for k, v in self.id_to_symbols_dict.items()])
        y = [map_dict[xx.upper()] for xx in x]
        if len(y) == 1:
            y = y[0]
        return y

    @property
    def strippednames_to_id_dict(self):
        return dict(zip([i.split(".")[0] for i in self.genome_tab[KEY_SYMBOL]],
                        self.genome_tab[KEY_ID].values.tolist()))


class CustomFeatureContainer(GenomeContainer):

    def __init__(
            self,
            genome_tab: pandas.DataFrame,
            organism: str,
    ):
        """

        :param genome_tab: Table characterising feature space. Must be a data frame with 3 columns:

            - "gene_name": Name of features.
            - "gene_id": ID of features, can be the same as values of "gene_name"
            - "gene_biotype": Types of features, can be arbitrary like "embedding"
        """
        self.assembly = "custom"
        assert len(genome_tab.columns) == 3
        assert KEY_SYMBOL in genome_tab.columns
        assert KEY_ID in genome_tab.columns
        assert KEY_TYPE in genome_tab.columns
        self.genome_tab = genome_tab
        self._organism = organism

    @property
    def organism(self):
        return self._organism


class ReactiveFeatureContainer(GenomeContainerBase):

    """
    Container class for features that can reactively loaded based on features present in data.

    The symbols are added by the store that uses this container.
    """

    def __init__(self, **kwargs):
        self._symbols = None

    @property
    def organism(self):
        return None

    def set(
            self,
            symbols: Union[None, str, List[str]] = None,
    ):
        """
        Set feature space to identifiers (symbol or ensemble ID).

        Note that there is no background (assembly) feature space to subset in this class.

        :param symbols: Gene symbol(s) of gene(s) to subset genome to. Elements have to appear in genome.
            Separate in string via "," if choosing multiple or supply as list of string.
        """
        self._symbols = symbols

    @property
    def symbols(self) -> List[str]:
        return self._symbols

    @symbols.setter
    def symbols(self, x):
        self._symbols = x

    @property
    def ensembl(self) -> List[str]:
        return self._symbols

    @property
    def biotype(self) -> List[str]:
        return ["custom" for x in self._symbols]

    @property
    def n_var(self) -> int:
        return len(self.symbols) if self.symbols is not None else None



import logging
import numpy as np
import os

import sys
from rich import print

from sfaira.commands.utils import get_ds
from sfaira.consts.utils import clean_doi

try:
    import sfaira_extension as sfairae
except ImportError:
    sfairae = None

log = logging.getLogger(__name__)


class DataloaderTester:

    def __init__(self, path_loader, path_data, doi):
        self.WD = os.path.dirname(__file__)
        self.path_loader = path_loader
        self.path_data = path_data
        self.cwd = os.getcwd()
        self.doi = doi
        self.doi_sfaira_repr = ''

    def test_dataloader(self, clean_tsvs: bool, in_phase_3: bool):
        """
        Runs a predefined unit test on a given dataloader.
        """
        self.doi_sfaira_repr = clean_doi(self.doi)
        self._test_dataloader(clean_tsvs=clean_tsvs, in_phase_3=in_phase_3)

    def _get_ds(self):
        return get_ds(doi_sfaira_repr=self.doi_sfaira_repr, path_data=self.path_data, path_loader=self.path_loader)

    def _test_dataloader(self, clean_tsvs: bool, in_phase_3: bool):
        """
        Tests the dataloader.
        """
        print('[bold blue]Conflicts are not automatically resolved.')
        print('[bold blue]In case of coflicts, please go back to [bold]https://www.ebi.ac.uk/ols/ontologies/cl[blue] '
              'and add the correct cell ontology class name into the .tsv "target" column.')

        ds, cache_path = self._get_ds()
        if clean_tsvs:
            ds.clean_ontology_class_maps()

        ds, cache_path = self._get_ds()
        ds.load(load_raw=True, allow_caching=False)
        # Test that not too much of the count matrix is lost during feature streamlining:
        # This would indicate an issue with the feature assignments.
        for k, v in ds.datasets.items():
            signal_raw = np.asarray(v.adata.X.sum()).sum()
            v.streamline_features(schema="cellxgene:" + "2.0.0")
            signal_proc = np.asarray(v.adata.X.sum()).sum()
            if signal_proc < 0.01 * signal_raw and v.feature_type != "peak":
                print('[bold red]Mapping your feature space to a reference annotation resulted in a heavy loss of '
                      f'counts in dataset {k}.')
                print(f'[bold red]From {signal_raw} total counts before streamlining, {signal_proc} are left after.')
                print('[bold red]Consider revising feature meta data.')
                sys.exit()
            v.streamline_metadata(schema="cellxgene")
        print("[bold blue]Completed testing of data loader, the data loader is now ready for use.")
        if in_phase_3:
            print('[bold orange]Sfaira butler: "You data loader is finished!"')
            print('[bold orange]               "Proceed to phase 4 (publish) or use data loader."')
            print('[bold orange]               "Copy the following lines as a post into the pull request:"')
            print('[bold blue]=========================')
            print('[bold blue]Data loader test passed:')
            for x in ds.datasets.keys():
                print(f'[bold blue]    - {x}')
            print('[bold blue]=========================')
        else:
            print('[bold orange]Sfaira butler: "You data loader works!"')



from sfaira.data.dataloaders.loaders import DatasetSuperGroupLoaders

# TODO export this into a maintenance module.
# def test_map_celltype_to_ontology():
#    # Paths do not matter here as data sets are not loaded for these operations.
#    dsgl = DatasetSuperGroupLoaders(data_path="", meta_path="", cache_path="")
#    for x in dsgl.dataset_groups:
#        x.clean_ontology_class_map()



from typing import Dict

import anndata
import numpy as np

from sfaira.data.store.carts.base import CartBase
from sfaira.data.store.carts.single import CartSingle


class CartMulti(CartBase):

    """
    Cart for a DistributedStoreMultipleFeatureSpaceBase().
    """

    carts: Dict[str, CartSingle]
    intercalated: bool

    def __init__(self, carts: Dict[str, CartSingle], intercalated: bool = False):
        """

        :param carts: The generators to combine.
        :param intercalated: Whether to intercalate batches from both generators. Intercalates at frequency such that
            all generators terminate roughly at the same time.
            time.
        """
        self.carts = carts
        self.intercalated = intercalated
        self._ratios = None

    @property
    def adata(self) -> Dict[str, anndata.AnnData]:
        """
        Assembles a dictionary of slices of this cart based on .obs_idx as anndata instances per organism.
        """
        return dict([(k, v.adata) for k, v in self.carts.items()])

    def iterator(self, repeat: int = 1, shuffle_buffer: int = 0):
        keep_repeating = True
        num_repetitions = 0

        if self.intercalated:
            while keep_repeating:
                ratios = self.ratios.copy()
                print(f"GENERATOR: intercalating generators at ratios {ratios}")
                # Document which generators are still yielding batches:
                yielding = np.ones((ratios.shape[0],)) == 1.
                iterators = [v.iterator(repeat=1, shuffle_buffer=shuffle_buffer) for v in self.carts.values()]
                while np.any(yielding):
                    # Loop over one iterator length adjusted cycle of emissions.
                    for i, (gi, n) in enumerate(zip(iterators, ratios)):
                        n_rounded = int(n) + np.random.binomial(n=1, p=n - int(n))
                        for _ in range(n_rounded):
                            try:
                                x = next(gi)
                                yield x
                            except StopIteration:
                                yielding[i] = False
                num_repetitions += 1
                keep_repeating = (num_repetitions < repeat) or (repeat <= 0)
        else:
            while keep_repeating:
                for gi in self.carts.values():
                    for x in gi.iterator():
                        yield x

                num_repetitions += 1
                keep_repeating = (num_repetitions < repeat) or (repeat <= 0)

    @property
    def n_batches(self) -> int:
        return int(np.sum([v.n_batches for v in self.carts.values()]))

    @property
    def n_obs(self) -> int:
        """Total number of observations in cart."""
        return int(np.sum([v.n_obs for v in self.carts.values()]))

    @property
    def n_obs_selected(self) -> int:
        """Total number of selected observations in cart."""
        return int(np.sum([v.n_obs_selected for v in self.carts.values()]))

    @property
    def n_var(self) -> Dict[str, int]:
        """Total number of features defined for return in cart."""
        return dict([(k, v.n_var) for k, v in self.carts.items()])

    @property
    def obs(self):
        return dict([(k, v.obs) for k, v in self.carts.items()])

    @property
    def obs_idx(self):
        """
        Dictionary of integer observation indices to select from cart. These will be emitted if data is queried.
        """
        return dict([(k, v.obs_idx) for k, v in self.carts.items()])

    @obs_idx.setter
    def obs_idx(self, x):
        if x is None:
            x = dict([(k, None) for k in self.carts.keys()])
        for k in self.carts.keys():
            assert k in x.keys(), (x.keys(), self.carts.keys())
            self.carts[k].obs_idx = x[k]
        self._ratios = None  # Reset ratios.

    def move_to_memory(self):
        """
        Persist underlying arrays into memory in sparse.COO format.
        """
        for v in self.carts.values():
            v.move_to_memory()

    @property
    def ratios(self):
        """
        Define relative drawing frequencies from iterators for intercalation.

        Note that these can be float and will be randomly rounded during intercalation.
        """
        if self._ratios is None:
            gen_lens = np.array([v.n_batches for v in self.carts.values()])  # eg. [10, 15, 20]
            # Compute ratios of sampling so that one batch is drawn from the smallest generator per intercalation cycle.
            # See also self.iterator.
            freq = gen_lens / np.min(gen_lens)  # eg. [1.0, 1.5, 2.0]
            self._ratios = freq
        return self._ratios

    @property
    def var(self):
        return dict([(k, v.var) for k, v in self.carts.items()])

    @property
    def x(self):
        return dict([(k, v.x) for k, v in self.carts.items()])



import anndata
from typing import Union

from sfaira.data import DatasetBase


class DatasetInteractive(DatasetBase):

    def __init__(
            self,
            data: anndata.AnnData,
            feature_symbol_col: Union[str, None] = 'index',
            feature_id_col: Union[str, None] = None,
            feature_type_col: Union[str, None] = None,
            dataset_id: str = "interactive_dataset",
            data_path: Union[str, None] = ".",
            meta_path: Union[str, None] = ".",
            cache_path: Union[str, None] = ".",
    ):
        """
        Load data set into sfaira data format.

        :param data: Data set.
        :param feature_symbol_col: Column name in .var which contains gene symbols. Set to "index" to use the index.
        :param feature_id_col:  Column name in .var which contains ENSG symbols. Set to "index" to use the index.
        :param feature_type_col:  Column name in .var which contains feature type.
        :param dataset_id: Identifer of data set.
        :param data_path:
        :param meta_path:
        :param cache_path:
        """
        super().__init__(data_path=data_path, meta_path=meta_path, cache_path=cache_path)
        self.id = dataset_id

        self.author = "interactive_dataset"
        self.doi_journal = "interactive_dataset"
        self.download_url_data = "."
        self.download_url_meta = "."

        self.layer_processed = "X"

        self.feature_symbol_var_key = feature_symbol_col
        self.feature_id_var_key = feature_id_col
        self.feature_type_var_key = feature_type_col

        self.adata = data

    def _load(self):
        pass



import os
import numpy as np


def load(data_dir, **kwargs):
    import anndata2ri
    from rpy2.robjects import r

    fn = os.path.join(data_dir, "tissue.rdata")
    anndata2ri.activate()  # TODO: remove global activation of anndata2ri and use localconverter once it's fixed
    adata = r(
        f"library(Seurat)\n"
        f"load('{fn}')\n"
        f"new_obj = CreateSeuratObject(counts = tissue@raw.data)\n"
        f"new_obj@meta.data = tissue@meta.data\n"
        f"as.SingleCellExperiment(new_obj)\n"
    )
    adata.obs["nGene"] = adata.obs["nGene"].astype(np.int32)

    return adata



try:
    import tensorflow as tf
except ImportError:
    tf = None


class PreprocInput(tf.keras.layers.Layer):

    def __init__(
            self,
            name='preproc_input',
            **kwargs
    ):
        super().__init__(name=name, **kwargs)

    def call(self, inputs, **kwargs):
        sf = tf.reduce_sum(inputs, axis=1, keepdims=True) / 10000.
        inputs = tf.divide(inputs, sf)
        inputs = tf.math.log(inputs + 1)
        return inputs



import numpy as np
import os
import pandas as pd
import pydoc
import pytest

from sfaira.data.dataloaders.loaders import DatasetSuperGroupLoaders

from sfaira.unit_tests.directories import DIR_DATA_LOADERS_CACHE

SYMBOL_COL = "target"
ID_COL = "target_id"
UNKNOWN_IDS = ["UNKNOWN", "NOT_A_CELL"]


def test_celltype_update():
    """
    Warning: changes files in package if ontology mismatches are found!
    """
    dssg = DatasetSuperGroupLoaders(data_path=DIR_DATA_LOADERS_CACHE, meta_path=DIR_DATA_LOADERS_CACHE,
                                    cache_path=DIR_DATA_LOADERS_CACHE)
    for dsg in dssg.dataset_groups:
        collection = dsg.collection_id
        for ds in dsg.datasets.values():
            path_dsg = os.path.dirname(str(pydoc.locate(f"sfaira.data.dataloaders.loaders.{collection}.FILE_PATH")))
            for f in os.listdir(path_dsg):
                if f.split(".")[-1] == "tsv":  # Narrow down to ontology map files.
                    fn_map = os.path.join(path_dsg, f)
                    if os.path.exists(fn_map):
                        # Access reading and value protection mechanisms from first data set loaded in group.
                        tab = ds._read_ontology_class_map(fn=fn_map)
                        tab[SYMBOL_COL] = [
                            ds.celltypes_universe.onto_cl.convert_to_name(x)
                            if x not in UNKNOWN_IDS else x
                            for x in tab[ID_COL].values
                        ]
                        tab.to_csv(path_or_buf=fn_map, index=False, sep="\t")
                    else:
                        print(f"did not find map for {collection}, {ds.fn_ontology_class_map_tsv}: {fn_map}")



import numpy as np
try:
    import tensorflow as tf
except ImportError:
    tf = None


def custom_mse(y_true, y_pred, sample_weight=None):
    y_pred = tf.split(y_pred, num_or_size_splits=2, axis=1)[0]
    se = tf.square(tf.subtract(y_true, y_pred))
    se_red = tf.reduce_mean(se)
    return se_red


def custom_mean_squared_logp1_error(y_true, y_pred, sample_weight=None):
    y_pred = tf.split(y_pred, num_or_size_splits=2, axis=1)[0]
    y_true = tf.math.log(y_true + 1.)
    y_pred = tf.math.log(y_pred + 1.)
    se = tf.square(tf.subtract(y_true, y_pred))
    se_red = tf.reduce_mean(se)
    return se_red


def custom_negll_nb(y_true, y_pred, sample_weight=None):
    x = y_true
    loc, scale = tf.split(y_pred, num_or_size_splits=2, axis=1)

    eta_loc = tf.math.log(loc)
    eta_scale = tf.math.log(scale)

    log_r_plus_mu = tf.math.log(scale + loc)

    ll = tf.math.lgamma(scale + x)
    ll = ll - tf.math.lgamma(x + tf.ones_like(x))
    ll = ll - tf.math.lgamma(scale)
    ll = ll + tf.multiply(x, eta_loc - log_r_plus_mu) + tf.multiply(scale, eta_scale - log_r_plus_mu)

    ll = tf.clip_by_value(ll, -300, 300, "log_probs")
    neg_ll = -ll
    neg_ll = tf.reduce_mean(neg_ll)
    return neg_ll


def custom_negll_gaussian(y_true, y_pred, sample_weight=None):
    loc, scale = tf.split(y_pred, num_or_size_splits=2, axis=1)

    ll = -tf.math.log(scale * tf.math.sqrt(2. * np.pi)) - 0.5 * tf.math.square((y_true - loc) / scale)
    ll = tf.clip_by_value(ll, -300, 300, "log_probs")
    neg_ll = -ll
    neg_ll = tf.reduce_mean(neg_ll)
    return neg_ll


def custom_kl(y_true, y_pred, sample_weight=None):
    expected_logqz_x, expected_logpz = tf.split(y_pred, num_or_size_splits=2, axis=1)
    kl_loss = tf.reduce_mean(expected_logqz_x - expected_logpz)
    return kl_loss


def custom_cce_agg(y_true, y_pred, sample_weight=None):
    y_pred = tf.clip_by_value(y_pred, 1e-10, 1., "y_pred")
    ll_cce_agg = -tf.math.log(tf.reduce_mean(y_true * y_pred, axis=1, keepdims=False))
    return ll_cce_agg


class CustomAccAgg(tf.keras.metrics.Metric):

    def __init__(self, name='acc_agg', **kwargs):
        super(CustomAccAgg, self).__init__(name=name, **kwargs)
        self.acc_agg = self.add_weight(name='acc', initializer='zeros')
        self.count = self.add_weight(name='count', initializer='zeros')

    def update_state(self, y_true, y_pred, sample_weight=None):
        phat_pos_agg = tf.reduce_sum(y_true * y_pred, axis=1, keepdims=True)
        acc_agg = tf.cast(
            phat_pos_agg > tf.reduce_max((tf.ones_like(y_true) - y_true) * y_pred, axis=1),
            dtype=y_true.dtype
        )
        # Do not use weighting for accuracy.
        self.acc_agg.assign_add(tf.reduce_mean(acc_agg))
        self.count.assign_add(1.)

    def result(self):
        return tf.divide(self.acc_agg, self.count)

    def reset_states(self):
        self.acc_agg.assign(0.)
        self.count.assign(0.)


class CustomTprClasswise(tf.keras.metrics.Metric):

    def __init__(self, k: int, name='tpr', **kwargs):
        super(CustomTprClasswise, self).__init__(name=name, **kwargs)
        self.tp = self.add_weight(shape=(k,), name='tp', initializer='zeros')
        self.fn = self.add_weight(shape=(k,), name='fn', initializer='zeros')

    def update_state(self, y_true, y_pred, sample_weight=None):
        tp_by_class = tf.reduce_sum(tf.cast(
            y_pred == tf.reduce_max(y_pred, axis=1, keepdims=True),
            dtype=y_true.dtype
        ) * y_true, axis=0)
        fn_by_class = tf.reduce_sum(tf.cast(
            y_pred < tf.reduce_max(y_pred, axis=1, keepdims=True),
            dtype=y_true.dtype
        ) * y_true, axis=0)
        self.tp.assign_add(tp_by_class)
        self.fn.assign_add(fn_by_class)

    def result(self):
        # Catch division by zero, in that case tpr becomes zero after clipping the divisor to 1.
        divisor = tf.clip_by_value(self.tp + self.fn, 1, np.inf, "divisor")
        tpr = tf.reduce_mean(self.tp / divisor)
        return tpr

    def reset_states(self):
        self.tp.assign(tf.zeros_like(self.tp))
        self.fn.assign(tf.zeros_like(self.fn))


class CustomFprClasswise(tf.keras.metrics.Metric):

    def __init__(self, k: int, name='fpr', **kwargs):
        super(CustomFprClasswise, self).__init__(name=name, **kwargs)
        self.fp = self.add_weight(shape=(k,), name='fp', initializer='zeros')
        self.tn = self.add_weight(shape=(k,), name='tn', initializer='zeros')

    def update_state(self, y_true, y_pred, sample_weight=None):
        fp_by_class = tf.reduce_sum(tf.cast(
            y_pred == tf.reduce_max(y_pred, axis=1, keepdims=True),
            dtype=y_true.dtype
        ) * (1. - y_true), axis=0)
        tn_by_class = tf.reduce_sum(tf.cast(
            y_pred < tf.reduce_max(y_pred, axis=1, keepdims=True),
            dtype=y_true.dtype
        ) * (1. - y_true), axis=0)
        self.fp.assign_add(fp_by_class)
        self.tn.assign_add(tn_by_class)

    def result(self):
        # Catch division by zero, in that case fpr becomes zero after clipping the divisor to 1.
        divisor = tf.clip_by_value(self.fp + self.tn, 1, np.inf, "divisor")
        fpr = tf.reduce_mean(self.fp / divisor)
        return fpr

    def reset_states(self):
        self.fp.assign(tf.zeros_like(self.fp))
        self.tn.assign(tf.zeros_like(self.tn))


class CustomF1Classwise(tf.keras.metrics.Metric):

    def __init__(self, k: int, name='f1', **kwargs):
        super(CustomF1Classwise, self).__init__(name=name, **kwargs)
        self.tp = self.add_weight(shape=(k,), name='tp', initializer='zeros')
        self.fp = self.add_weight(shape=(k,), name='fp', initializer='zeros')
        self.fn = self.add_weight(shape=(k,), name='fn', initializer='zeros')

    def update_state(self, y_true, y_pred, sample_weight=None):
        tp_by_class = tf.reduce_sum(tf.cast(
            y_pred == tf.reduce_max(y_pred, axis=1, keepdims=True),
            dtype=y_true.dtype
        ) * y_true, axis=0)
        fp_by_class = tf.reduce_sum(tf.cast(
            y_pred == tf.reduce_max(y_pred, axis=1, keepdims=True),
            dtype=y_true.dtype
        ) * (1. - y_true), axis=0)
        fn_by_class = tf.reduce_sum(tf.cast(
            y_pred < tf.reduce_max(y_pred, axis=1, keepdims=True),
            dtype=y_true.dtype
        ) * y_true, axis=0)
        self.tp.assign_add(tp_by_class)
        self.fp.assign_add(fp_by_class)
        self.fn.assign_add(fn_by_class)

    def result(self):
        # Catch divisions by zero, in that case precision or recall become zero after clipping the divisor to 1.
        divisor_precision = tf.clip_by_value(self.tp + self.fp, 1, np.inf, "divisor")
        divisor_recall = tf.clip_by_value(self.tp + self.fn, 1, np.inf, "divisor")
        precision = self.tp / divisor_precision
        recall = self.tp / divisor_recall
        precision = tf.clip_by_value(precision, 1e-100, np.inf, "divisor")
        recall = tf.clip_by_value(recall, 1e-100, np.inf, "divisor")
        f1 = tf.reduce_mean(2 * 1 / (1 / precision + 1 / recall))
        return f1

    def reset_states(self):
        self.tp.assign(tf.zeros_like(self.tp))
        self.fp.assign(tf.zeros_like(self.fp))
        self.fn.assign(tf.zeros_like(self.fn))



from typing import Union

from sfaira.versions.genomes.genomes import GenomeContainer, ReactiveFeatureContainer


class TopologyContainer:

    """
    Class interface for a YAML-style defined model topology that loads a genome container tailored to the model.
    """

    def __init__(
            self,
            topology: dict,
            topology_id: str,
            custom_genome_constainer: Union[GenomeContainer, None] = None,
    ):
        self.topology = topology
        if custom_genome_constainer is None:
            if self.topology["input"]["genome"] is not None:
                self.gc = GenomeContainer(
                    organism=" ".join(self.topology["input"]["genome"].split(".")[0].split("_")),
                    release=self.topology["input"]["genome"].split(".")[-1],
                )
            else:
                self.gc = ReactiveFeatureContainer()
        else:
            assert isinstance(custom_genome_constainer, GenomeContainer)
            self.gc = custom_genome_constainer
        self.gc.set(**dict([tuple(self.topology["input"]["genes"])]))
        self.topology_id = topology_id

    @property
    def model_type(self):
        return self.topology["model_type"]

    @property
    def output(self):
        return self.topology["output"]

    @property
    def n_var(self):
        return self.gc.n_var

    @property
    def organism(self):
        return self.gc.organism



LINEAR_TOPOLOGIES = {
    "0.1": {
        "model_type": "linear",
        "input": {
            "genome": "Mus_musculus.GRCm38.102",
            "genes": ["biotype", "protein_coding"],
        },
        "output": {},
        "hyper_parameters": {
            "latent_dim": 64,
            "l1_coef": 0.,
            "l2_coef": 0.,
            "positive_components": False,
            "output_layer": "nb_shared_disp"
        }
    },

    "0.2": {
        "model_type": "linear",
        "input": {
            "genome": "Mus_musculus.GRCm38.102",
            "genes": ["biotype", "protein_coding"],
        },
        "output": {},
        "hyper_parameters": {
            "latent_dim": 128,
            "l1_coef": 0.,
            "l2_coef": 0.,
            "positive_components": False,
            "output_layer": "nb_shared_disp"
        }
    },

    "0.3": {
        "model_type": "linear",
        "input": {
            "genome": "Mus_musculus.GRCm38.102",
            "genes": ["biotype", "protein_coding"],
        },
        "output": {},
        "hyper_parameters": {
            "latent_dim": 128,
            "l1_coef": 0.,
            "l2_coef": 0.,
            "positive_components": False,
            "output_layer": "nb_const_disp"
        }
    }
}

# Load versions from extension if available:
try:
    from sfaira_extension.versions.topology_versions.mouse.embedding import LINEAR_TOPOLOGIES as LINEAR_TOPOLOGIES_EXTENSION
    LINEAR_TOPOLOGIES = {
        **LINEAR_TOPOLOGIES,
        **LINEAR_TOPOLOGIES_EXTENSION
    }
except ImportError:
    pass



from sfaira.models.celltype.base import BasicModelKerasCelltype
from sfaira.models.celltype.marker import CellTypeMarker, CellTypeMarkerVersioned
from sfaira.models.celltype.mlp import CellTypeMlp, CellTypeMlpVersioned



import anndata
import numpy as np
import pandas as pd
import scanpy as sc
import scipy.sparse
from typing import Union

from sfaira.consts.adata_fields import AdataIdsCellxgene

DEFAULT_CELLXGENE_VERSION = "2_0_0"


def cellxgene_export_adaptor(adata: anndata.AnnData, adata_ids: AdataIdsCellxgene, version: Union[None, str],
                             **kwargs) -> anndata.AnnData:
    """
    Projects a streamlined data set to the export-ready cellxgene format.
    """
    if version is None:
        version = DEFAULT_CELLXGENE_VERSION
    if version == "2_0_0":
        return cellxgene_export_adaptor_2_0_0(adata=adata, adata_ids=adata_ids, **kwargs)
    else:
        raise ValueError(f"Did not recognise cellxgene schema version {version}")


def cellxgene_export_adaptor_1_1_0(adata: anndata.AnnData, adata_ids: AdataIdsCellxgene, **kwargs) -> anndata.AnnData:
    """
    Cellxgene-schema 1.1.0
    """
    # Check input object characteristics:
    cellxgene_cols = [getattr(adata_ids, x) for x in adata_ids.ontology_constrained] + \
                     [getattr(adata_ids, x) for x in adata_ids.obs_keys
                      if x not in adata_ids.ontology_constrained] + \
                     [getattr(adata_ids, x) + adata_ids.onto_id_suffix
                      for x in adata_ids.ontology_constrained]
    for x in cellxgene_cols:
        if x not in adata.obs.columns:
            raise ValueError(f"Cannot streamlined data set {adata.uns['id']} to cellxgene format because meta data {x} "
                             f"is missing and the corresponding .obs column could not be written.\n"
                             f"Columns found were {adata.obs.columns}.")
    # 1) Modify .uns
    adata.uns["layer_descriptions"] = {"X": "raw"}
    adata.uns["version"] = {
        "corpora_encoding_version": "0.1.0",
        "corpora_schema_version": "1.1.0",
    }
    adata.uns["contributors"] = {
        "name": "sfaira",
        "email": "https://github.com/theislab/sfaira/issues",
        "institution": "sfaira",
    }
    # TODO port this into organism ontology handling.
    # Infer organism from adata object.
    organism = np.unique(adata.obs[adata_ids.organism].values)[0]
    if organism == "musmusculus":
        adata.uns["organism"] = "Mus musculus"
        adata.uns["organism_ontology_term_id"] = "NCBITaxon:10090"
    elif organism == "homosapiens":
        adata.uns["organism"] = "Homo sapiens"
        adata.uns["organism_ontology_term_id"] = "NCBITaxon:9606"
    else:
        raise ValueError(f"organism {organism} currently not supported by cellxgene schema")
    # 2) Modify .obs
    # Correct unknown cell type entries:
    adata.obs[adata_ids.cell_type] = [
        x if x not in [adata_ids.unknown_metadata_identifier, adata_ids.not_a_cell_celltype_identifier]
        else "native cell" for x in adata.obs[adata_ids.cell_type]]
    adata.obs[adata_ids.cell_type + adata_ids.onto_id_suffix] = [
        x if x not in [adata_ids.unknown_metadata_identifier, adata_ids.not_a_cell_celltype_identifier]
        else "CL:0000003" for x in adata.obs[adata_ids.cell_type + adata_ids.onto_id_suffix]]
    # Reorder data frame to put ontology columns first:
    adata.obs = adata.obs[cellxgene_cols + [x for x in adata.obs.columns if x not in cellxgene_cols]]
    # 3) Modify .X
    # Check if .X is counts: The conversion are based on the assumption that .X is csr.
    assert isinstance(adata.X, scipy.sparse.csr_matrix), type(adata.X)
    count_values = np.unique(np.asarray(adata.X.todense()))
    if not np.all(count_values % 1. == 0.):
        print(f"WARNING: not all count entries were counts, "
              f"the maximum deviation from integer is "
              f"{np.max([x % 1. if x % 1. < 0.5 else 1. - x % 1. for x in count_values])}. "
              f"The count matrix is rounded.")
        adata.X.data = np.rint(adata.X.data)
    return adata


def cellxgene_export_adaptor_2_0_0(adata: anndata.AnnData, adata_ids: AdataIdsCellxgene, obs_keys_batch,
                                   mask_portal_fields: bool = True, **kwargs) -> anndata.AnnData:
    """
    Cellxgene-schema 2.0.0.

    Documented here: https://github.com/chanzuckerberg/single-cell-curation/blob/main/schema/2.0.3/schema.md
    """
    obs_keys_autofill = [getattr(adata_ids, x) for x in adata_ids.ontology_constrained]
    uns_keys_autofill = []
    var_keys_autofill = [adata_ids.feature_symbol, adata_ids.feature_reference]
    raw_var_keys_remove = [adata_ids.feature_is_filtered]
    # 1) Modify .uns
    adata.uns["X_normalization"] = "log_normalized"
    adata.uns["X_approximate_distribution"] = "normal"
    adata.uns["schema_version"] = "2.0.0"
    adata.uns[adata_ids.author] = {
        "name": "sfaira",
        "email": "https://github.com/theislab/sfaira/issues",
        "institution": "sfaira",
    }
    if obs_keys_batch is not None:
        adata.uns["batch_condition"] = obs_keys_batch.split("*")
    # 2) Modify .obs
    # a) Correct unknown cell type entries:
    adata.obs[adata_ids.cell_type] = [
        x if x not in [adata_ids.unknown_metadata_identifier, adata_ids.not_a_cell_celltype_identifier]
        else "native cell" for x in adata.obs[adata_ids.cell_type]]
    adata.obs[adata_ids.cell_type + adata_ids.onto_id_suffix] = [
        x if x not in [adata_ids.unknown_metadata_identifier, adata_ids.not_a_cell_celltype_identifier]
        else "CL:0000003" for x in adata.obs[adata_ids.cell_type + adata_ids.onto_id_suffix]]
    # b) Correct unknown disease entries:
    adata.obs[adata_ids.disease] = [
        x if x not in [adata_ids.unknown_metadata_identifier] else "healthy"
        for x in adata.obs[adata_ids.disease]]
    adata.obs[adata_ids.disease + adata_ids.onto_id_suffix] = [
        x if x not in [adata_ids.unknown_metadata_identifier] else "PATO:0000461"
        for x in adata.obs[adata_ids.disease + adata_ids.onto_id_suffix]]
    # Reorder data frame to put ontology columns first:
    cellxgene_cols = [getattr(adata_ids, x) for x in adata_ids.ontology_constrained] + \
                     [getattr(adata_ids, x) for x in adata_ids.obs_keys
                      if x not in adata_ids.ontology_constrained] + \
                     [getattr(adata_ids, x) + adata_ids.onto_id_suffix
                      for x in adata_ids.ontology_constrained]
    adata.obs = adata.obs[cellxgene_cols + [x for x in adata.obs.columns if x not in cellxgene_cols]]
    # 3) Modify .X, assuming that .X is raw (TODO)
    # Check if .X is counts: The conversion are based on the assumption that .X is csr.
    assert isinstance(adata.X, scipy.sparse.csr_matrix), type(adata.X)
    count_values = np.unique(np.asarray(adata.X.todense()))
    if not np.all(count_values % 1. == 0.):
        print(f"WARNING: not all count entries were counts, "
              f"the maximum deviation from integer is "
              f"{np.max([x % 1. if x % 1. < 0.5 else 1. - x % 1. for x in count_values])}. "
              f"The count matrix is rounded.")
        adata.X.data = np.rint(adata.X.data)
    # 4) Modify .var:
    adata.var[adata_ids.feature_biotype] = "gene"
    adata.var[adata_ids.feature_reference] = adata.obs[adata_ids.organism + adata_ids.onto_id_suffix].values[0]
    # TODO set filtered based on feature overlap between raw and X.
    adata.var[adata_ids.feature_is_filtered] = False
    adata.var[adata_ids.feature_biotype] = pd.Categorical(adata.var[adata_ids.feature_biotype].values.tolist())
    adata.var.index = adata.var[adata_ids.feature_id].values
    # Modify ensembl ID writing:
    # adata.var[adata_ids.feature_id] = ["G:".join(x.split("G")) for x in adata.var[adata_ids.feature_id]]
    # adata.var.index = ["G:".join(x.split("G")) for x in adata.var.index]
    # 5) Take out elements that are auto-filled by cellxgene upload interface:
    if mask_portal_fields:
        for k in obs_keys_autofill:
            del adata.obs[k]
        for k in uns_keys_autofill:
            del adata.uns[k]
        for k in var_keys_autofill:
            del adata.var[k]
    # 6) Modify .raw, assuming that .X is raw: TODO below
    # Note that this depends on var cleaning in step 5.
    var_raw = pd.DataFrame(dict([(k, v) for k, v in adata.var.items() if k not in raw_var_keys_remove]))
    adata.raw = anndata.AnnData(adata.X, var=var_raw)
    # Log-normalise values in .X TODO make optional
    sc.pp.normalize_per_cell(adata)
    sc.pp.log1p(adata)
    # 7) Check if default embedding is present, add in otherwise:
    # First check if any pre-computed embedding is given.
    # If that is not the case, compute a default UMAP.
    # Define hierarchy of embeddings accepted as defaults, first one matched will be chosen:
    default_embedding_names = ["X_umap", "X_tsne", "X_draw_graph_fa"]
    if adata.uns[adata_ids.default_embedding] == adata_ids.unknown_metadata_identifier:
        found_default = False
        counter = 0
        while not found_default and counter < len(default_embedding_names):
            if default_embedding_names[counter] in adata.obsm.keys():
                adata.uns[adata_ids.default_embedding] = default_embedding_names[counter]
                found_default = True
            counter += 1
        if not found_default and adata.n_obs > 10:
            adata_embedding = adata.raw.to_adata().copy()
            sc.pp.normalize_per_cell(adata_embedding)
            sc.pp.log1p(adata_embedding)
            sc.pp.pca(adata_embedding)
            sc.pp.neighbors(adata_embedding)
            sc.tl.umap(adata_embedding)
            adata.obsm["X_umap"] = adata_embedding.obsm["X_umap"]
            adata.uns[adata_ids.default_embedding] = "X_umap"
    return adata



import anndata
import os

from sfaira.data import DatasetBase


class Dataset(DatasetBase):

    """
    TODO data link is outdated. Maybe update to this
    https://data.humancellatlas.org/explore/projects/cc95ff89-2e68-4a08-a234-480eca21ce79/project-matrices.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.download_url_data = "https://data.humancellatlas.org/project-assets/project-matrices/" \
                                 "cc95ff89-2e68-4a08-a234-480eca21ce79.homo_sapiens.loom"
        self.download_url_meta = None

        self.assay_sc = "10x 3' v2"
        self.author = "Regev"
        self.disease = "healthy"
        self.doi_journal = "no_doi_regev"
        self.layer_counts = "X"
        self.organ_obs_key = "derived_organ_parts_label"
        self.organism = "Homo sapiens"
        self.primary_data = True
        self.sample_source = "primary_tissue"
        self.year = 2018

        self.feature_id_var_key = "Accession"
        self.feature_symbol_var_key = "index"
        self.feature_type = "rna"

        self.set_dataset_id(idx=1)


def load(data_dir, **kwargs):
    fn = os.path.join(data_dir, "cc95ff89-2e68-4a08-a234-480eca21ce79.homo_sapiens.loom")
    adata = anndata.read_loom(fn)
    adata = adata[adata.obs["emptydrops_is_cell"] == "t"].copy()

    return adata



import os
import gzip
import anndata
import shutil


def load(data_dir, **kwargs):

    sex_dict = {
        "F": "female",
        "M": "male"
    }

    dev_stage_dict = {
        72: "11th week post-fertilization homosapiens stage",
        74: "11th week post-fertilization homosapiens stage",
        85: "13th week post-fertilization homosapiens stage",
        89: "13th week post-fertilization homosapiens stage",
        90: "13th week post-fertilization homosapiens stage",
        94: "14th week post-fertilization homosapiens stage",
        96: "14th week post-fertilization homosapiens stage",
        100: "15th week post-fertilization homosapiens stage",
        110: "16th week post-fertilization homosapiens stage",
        112: "17th week post-fertilization homosapiens stage",
        113: "17th week post-fertilization homosapiens stage",
        115: "17th week post-fertilization homosapiens stage",
        117: "17th week post-fertilization homosapiens stage",
        119: "18th week post-fertilization homosapiens stage",
        120: "18th week post-fertilization homosapiens stage",
        122: "18th week post-fertilization homosapiens stage",
        125: "18th week post-fertilization homosapiens stage",
        129: "19th week post-fertilization homosapiens stage",
    }

    organ_dict = {
        "Adrenal": "adrenal gland",
        "Cerebellum": "cerebellum",
        "Cerebrum": "telencephalon",
        "Eye": "eye",
        "Heart": "heart",
        "Intestine": "intestine",
        "Kidney": "kidney",
        "Liver": "liver",
        "Lung": "lung",
        "Muscle": "muscle organ",
        "Pancreas": "pancreas",
        "Placenta": "placenta",
        "Spleen": "spleen",
        "Stomach": "stomach",
        "Thymus": "thymus",
    }

    fn = os.path.join(data_dir, "GSE156793_S3_gene_count.loom.gz")
    fn_tmp = os.path.join(os.path.expanduser("~"), "tmp.loom")
    with gzip.open(fn, 'rb') as f_in:
        with open(fn_tmp, 'wb') as f_out:
            shutil.copyfileobj(f_in, f_out)
    adata = anndata.read_loom(fn_tmp)
    os.remove(fn_tmp)

    adata.obs["Sex"] = [sex_dict[x] for x in adata.obs["Sex"]]
    adata.obs["Organ"] = [organ_dict[x] for x in adata.obs["Organ"]]
    adata.obs["Developmental_stage"] = [dev_stage_dict[x] for x in adata.obs["Development_day"]]

    return adata



"""
The classes contains constants related to sfaira streamlined meta data files.
"""

META_DATA_FIELDS = {
    "annotated": bool,
    "author": str,
    "assay_sc": str,
    "assay_differentiation": str,
    "assay_type_differentiation": str,
    "bio_sample": str,
    "cell_line": str,
    "cellontology_class": str,
    "cellontology_id": str,
    "development_stage": str,
    "disease": str,
    "doi": str,
    "download_url_data": str,
    "download_url_meta": str,
    "id": str,
    "ncells": int,
    "normalization": str,
    "organ": str,
    "organism": str,
    "sample_source": str,
    "state_exact": str,
    "tech_sample": str,
    "year": int,
}



try:
    import tensorflow as tf
except ImportError:
    tf = None
from typing import Union

from sfaira.versions.metadata import CelltypeUniverse
from sfaira.versions.topologies import TopologyContainer
from sfaira.models.celltype.base import BasicModelKerasCelltype
from sfaira.models.pp_layer import PreprocInput


class LearnedThresholdLayer(tf.keras.layers.Layer):
    """
    A layer that thresholds the input with a learned threshold.
    """

    def __init__(
        self,
        out_dim,
        initializer="zeros",
        name='learned_threshold_layer',
        **kwargs
    ):
        super().__init__(name=name, **kwargs)
        self.bias = self.add_weight(
            "bias",
            shape=[1, out_dim],
            initializer=initializer
        )
        self.slope = self.add_weight(
            "slope",
            shape=[1, out_dim],
            initializer="zeros"
        )

    def call(self, inputs):
        x = inputs
        x = (x + self.bias) * tf.exp(self.slope)
        return tf.nn.sigmoid(x)


class CellTypeMarker(BasicModelKerasCelltype):
    """
    Marker gene-based cell type classifier: Learns whether or not each gene exceeds requires threshold
    and learns cell type assignment as linear combination of these marker gene presence probabilities.
    Activitiy and weight regularizers keep this sparse.
    """

    def __init__(
            self,
            in_dim: int,
            out_dim: int,
            l1_coef: float = 0.,
            l2_coef: float = 0.,
            kernel_initializer='glorot_uniform',
            bias_initializer='zeros',
            bias_regularizer=None,
            kernel_constraint=None,
            bias_constraint=None,
            name='celltype_marker',
            **kwargs
    ):
        """

        :param in_dim: Input feature dimension (number of features).
        :param out_dim: Number of output nodes: Number of cell types +1 for unkown cell type.
        :param l1_coef: L1 regularization hyperparameter on weights of last layer.
        :param l2_coef: L2 regularization hyperparameter on weights of last layer and activation of hidden.
        :param kernel_initializer: See tf.keras.layers.Dense() documentation.
        :param bias_initializer: See tf.keras.layers.Dense() documentation.
        :param kernel_regularizer: See tf.keras.layers.Dense() documentation.
        :param bias_regularizer: See tf.keras.layers.Dense() documentation.
        :param kernel_constraint: See tf.keras.layers.Dense() documentation.
        :param bias_constraint: See tf.keras.layers.Dense() documentation.
        """
        self.in_dim = in_dim
        self.out_dim = out_dim

        inputs = tf.keras.Input(shape=(in_dim,))
        x = PreprocInput()(inputs)
        # Expand one dimension to be able to use LocallyConnected1D.
        x = LearnedThresholdLayer(out_dim=self.in_dim)(x)
        y = tf.keras.layers.Dense(
            units=out_dim,
            activation="softmax",
            use_bias=True,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            kernel_regularizer=tf.keras.regularizers.l1_l2(l1=l1_coef, l2=l2_coef),
            bias_regularizer=bias_regularizer,
            activity_regularizer=None,
            kernel_constraint=kernel_constraint,
            bias_constraint=bias_constraint
        )(x)
        self.training_model = tf.keras.Model(inputs=[inputs, ], outputs=y, name=name)


class CellTypeMarkerVersioned(CellTypeMarker):

    def __init__(
            self,
            celltypes_version: CelltypeUniverse,
            topology_container: TopologyContainer,
            override_hyperpar: Union[dict, None] = None
    ):
        """

        :param celltypes_version:
        :param topology_container:
        :param override_hyperpar: Dictionary with hyper-parameters of model to override in preset hyper-parameter
            dictionary that is queried based on the topology_id. Can contain a subset of all hyperparameters.
        """
        # Get cell type version instance based on topology ID, organism and organ.
        hyperpar = topology_container.topology["hyper_parameters"]
        if override_hyperpar is not None:
            for k in list(override_hyperpar.keys()):
                hyperpar[k] = override_hyperpar[k]
        super().__init__(
            in_dim=topology_container.n_var,
            out_dim=celltypes_version.onto_cl.n_leaves,
            **hyperpar
        )
        print('passed hyperpar: \n', hyperpar)
        self._topology_id = topology_container.topology_id
        self.genome_size = topology_container.n_var
        self.model_class = "celltype"
        self.model_type = topology_container.model_type
        self.hyperparam = dict(
            list(hyperpar.items()) +  # noqa: W504
            [
                ("topology_id", self._topology_id),
                ("genome_size", self.genome_size),
                ("model_class", self.model_class),
                ("model_type", self.model_type),
                ("ntypes", celltypes_version.onto_cl.n_leaves),
            ]
        )



import anndata
import os
import scipy.sparse
import pandas as pd

disease_map = {"CASE": "pulmonary tuberculosis", "CONTROL": "healthy"}


def load(data_dir, sample_fn, **kwargs):
    fn = os.path.join(data_dir, sample_fn)
    adata = anndata.read_text(fn, delimiter="\t")
    adata = adata.T
    adata.X = scipy.sparse.csr_matrix(adata.X)
    metadata = pd.read_csv(os.path.join(data_dir, "GSE158769_meta_data.txt.gz"), sep="\t")
    adata.obs = adata.obs.join(metadata.set_index("cell_id"))
    adata.obs["cluster_name"] = adata.obs["cluster_name"].astype("str")

    adata.obs["disease"] = adata.obs["TB_status"].map(disease_map)

    adata.obsm["X_umap"] = adata.obs.loc[:, ["UMAP_1", "UMAP_2"]].values
    return adata



import anndata
import os
import scipy.sparse
import numpy as np


def load(data_dir, sample_fn, **kwargs):
    fn = os.path.join(data_dir, sample_fn)
    if sample_fn.split("_")[0] == "droplet":
        norm_const = 10000
        sf_key = "nUMI"
    else:
        norm_const = 1000000
        sf_key = "nReads"
    adata = anndata.read(fn)
    adata.X = scipy.sparse.csc_matrix(adata.X)
    adata.X = np.expm1(adata.X)
    adata.X = adata.X.multiply(scipy.sparse.csc_matrix(adata.obs[sf_key].values[:, None])).multiply(1 / norm_const)

    return adata



import numpy as np
import torch
import torchmetrics


def custom_mse(y_true, y_pred, sample_weight=None):
    y_pred = torch.chunk(y_pred, chunks=2, dim=1)[0]
    se = torch.square(torch.subtract(y_true, y_pred))
    se_red = torch.mean(se)
    return se_red


def custom_mean_squared_logp1_error(y_true, y_pred, sample_weight=None):
    y_pred = torch.chunk(y_pred, chunks=2, dim=1)[0]
    y_true = torch.log(y_true + 1.)
    y_pred = torch.log(y_pred + 1.)
    se = torch.square(torch.subtract(y_true, y_pred))
    se_red = torch.mean(se)
    return se_red


def custom_negll_nb(y_true, y_pred, sample_weight=None):
    x = y_true
    loc, scale = torch.chunk(y_pred, chunks=2, dim=1)

    eta_loc = torch.log(loc)
    eta_scale = torch.log(scale)

    log_r_plus_mu = torch.log(scale + loc)

    ll = torch.lgamma(scale + x)
    ll = ll - torch.lgamma(x + torch.ones_like(x))
    ll = ll - torch.lgamma(scale)
    ll = ll + torch.multiply(x, eta_loc - log_r_plus_mu) + torch.multiply(scale, eta_scale - log_r_plus_mu)

    ll = torch.clamp(ll, min=-300, max=300)
    neg_ll = -ll
    neg_ll = torch.mean(neg_ll)
    return neg_ll


def custom_negll_gaussian(y_true, y_pred, sample_weight=None):
    loc, scale = torch.chunk(y_pred, chunks=2, dim=1)

    ll = -torch.log(scale * torch.sqrt(2. * np.pi)) - 0.5 * torch.square((y_true - loc) / scale)
    ll = torch.clamp(ll, min=-300, max=300)
    neg_ll = -ll
    neg_ll = torch.mean(neg_ll)
    return neg_ll


def custom_kl(y_true, y_pred, sample_weight=None):
    expected_logqz_x, expected_logpz = torch.chunk(y_pred, chunks=2, dim=1)
    kl_loss = torch.mean(expected_logqz_x - expected_logpz)
    return kl_loss


def custom_cce_agg(y_true, y_pred, sample_weight=None):
    y_pred = torch.clamp(y_pred, min=1e-10, max=1.)
    ll_cce_agg = -torch.log(torch.mean(y_true * y_pred, dim=1, keepdim=False))
    return ll_cce_agg


class CustomAccAgg(torchmetrics.Metric):

    def __init__(self, dist_sync_on_step=False):
        super(CustomAccAgg, self).__init__(dist_sync_on_step=dist_sync_on_step)
        self.add_state('acc', default=torch.Tensor([0.]))
        self.add_state('count', default=torch.Tensor([0.]))

    def update(self, preds, target, sample_weight=None):
        phat_pos_agg = torch.sum(target * preds, dim=1, keepdim=True)
        acc_agg = (phat_pos_agg > torch.max((torch.ones_like(target) - target) * preds, dim=1)).type(target.dtype)
        # Do not use weighting for accuracy.
        self.acc += self.acc(torch.mean(acc_agg))
        self.count += 1.

    def compute(self):
        return torch.divide(self.acc, self.count)


class CustomTprClasswise(torchmetrics.Metric):

    def __init__(self, k: int, dist_sync_on_step=False):
        super(CustomTprClasswise, self).__init__(dist_sync_on_step=dist_sync_on_step)
        self.add_state('tp', default=torch.zeros(size=(k, )))
        self.add_state('fn', default=torch.zeros(size=(k, )))

    def update(self, preds, target, sample_weight=None):
        tp_by_class = torch.sum((preds == torch.max(preds, dim=1, keepdim=True)).type(target.dtype) * target, dim=0)
        fn_by_class = torch.sum((preds < torch.max(preds, dim=1, keepdim=True)).type(target.dtype) * target, dim=0)
        self.tp += tp_by_class
        self.fn += fn_by_class

    def compute(self):
        # Catch division by zero, in that case tpr becomes zero after clipping the divisor to 1.
        divisor = torch.clamp(self.tp + self.fn, min=1., max=np.inf)
        tpr = torch.mean(self.tp / divisor)
        return tpr


class CustomFprClasswise(torchmetrics.Metric):

    def __init__(self, k: int, dist_sync_on_step=False):
        super(CustomFprClasswise, self).__init__(dist_sync_on_step=dist_sync_on_step)
        self.add_state('fp', default=torch.zeros(size=(k, )))
        self.add_state('tn', default=torch.zeros(size=(k, )))

    def update(self, preds, target, sample_weight=None):
        fp_by_class = torch.sum(
            (preds == torch.max(preds, dim=1, keepdim=True)).type(target.dtype) * (1. - target),
            dim=0
        )
        tn_by_class = torch.sum(
            (preds < torch.max(preds, dim=1, keepdim=True)).type(target.dtype) * (1. - target),
            dim=0
        )
        self.fp += fp_by_class
        self.tn += tn_by_class

    def compute(self):
        # Catch division by zero, in that case fpr becomes zero after clipping the divisor to 1.
        divisor = torch.clamp(self.fp + self.tn, min=1., max=np.inf)
        fpr = torch.mean(self.fp / divisor)
        return fpr


class CustomF1Classwise(torchmetrics.Metric):

    def __init__(self, k: int, dist_sync_on_step=False):
        super(CustomF1Classwise, self).__init__(dist_sync_on_step=dist_sync_on_step)
        self.add_state('tp', default=torch.zeros(size=(k, )))
        self.add_state('fp', default=torch.zeros(size=(k, )))
        self.add_state('fn', default=torch.zeros(size=(k, )))

    def update(self, preds, target, sample_weight=None):
        tp_by_class = torch.sum((
            preds == torch.max(preds, dim=1, keepdim=True)).type(target.dtype) * target, dim=0)
        fp_by_class = torch.sum(
            (preds == torch.max(preds, dim=1, keepdim=True)).type(target.dtype) * (1. - target),
            dim=0
        )
        fn_by_class = torch.sum((preds < torch.max(preds, dim=1, keepdim=True)).type(target.dtype) * target, dim=0)
        self.tp.assign_add(tp_by_class)
        self.fp.assign_add(fp_by_class)
        self.fn.assign_add(fn_by_class)

    def compute(self):
        # Catch divisions by zero, in that case precision or recall become zero after clipping the divisor to 1.
        divisor_precision = torch.clamp(self.tp + self.fp, min=1, max=np.inf)
        divisor_recall = torch.clamp(self.tp + self.fn, min=1., max=np.inf)
        precision = self.tp / divisor_precision
        recall = self.tp / divisor_recall
        precision = torch.clamp(precision, min=1e-100, max=np.inf)
        recall = torch.clamp(recall, min=1e-100, max=np.inf)
        f1 = torch.mean(2 * 1 / (1 / precision + 1 / recall))
        return f1



import numpy as np
import os
import pytest

import sfaira.versions.genomes
from sfaira.consts import AdataIdsSfaira
from sfaira.data import DatasetSuperGroup, DatasetInteractive
from sfaira.data import Universe

from sfaira.unit_tests.data_for_tests.loaders import RELEASE_HUMAN, PrepareData
from sfaira.unit_tests.directories import DIR_TEMP, DIR_DATA_LOADERS_CACHE


def test_dsgs_instantiate():
    _ = Universe(data_path=DIR_DATA_LOADERS_CACHE, meta_path=DIR_DATA_LOADERS_CACHE, cache_path=DIR_DATA_LOADERS_CACHE)


def test_dsgs_crossref():
    """
    Tests if crossref attributes can be retrieved for all data loader entries with DOI journal defined.
    Attributes tested:
        - title
    """
    universe = Universe(data_path=DIR_DATA_LOADERS_CACHE, meta_path=DIR_DATA_LOADERS_CACHE,
                        cache_path=DIR_DATA_LOADERS_CACHE)
    for k, v in universe.datasets.items():
        title = v.title
        if title is None:
            if v.doi_journal is not None and "no_doi" not in v.doi_journal:
                raise ValueError(f"did not retrieve title for data set {k} with DOI: {v.doi_journal}.")


@pytest.mark.parametrize("organ", ["lung"])
def test_dsgs_subset_dataset_wise(organ: str):
    """
    Tests if subsetting results only in datasets of the desired characteristics.
    """
    ds = PrepareData().prepare_dsg(load=False)
    ds.subset(key="organism", values=["Homo sapiens"])
    ds.subset(key="organ", values=[organ])
    ds.load()
    for x in ds.dataset_groups:
        for k, v in x.datasets.items():
            assert v.organism == "Homo sapiens", v.organism
            assert v.ontology_container_sfaira.organ.is_a(query=v.organ, reference=organ), v.organ


def test_dsgs_config_write_load():
    fn = os.path.join(DIR_TEMP, "config.csv")
    ds = PrepareData().prepare_dsg(load=False)
    ds.subset(key="organism", values=["Homo sapiens"])
    ds.subset(key="organ", values=["lung"])
    ds.load()
    ds.write_config(fn=fn)
    ds2 = PrepareData().prepare_dsg()
    ds2.load_config(fn=fn)
    assert np.all(ds.ids == ds2.ids)


def test_dsgs_adata():
    ds = PrepareData().prepare_dsg(load=False)
    ds.subset(key="organism", values=["Homo sapiens"])
    ds.subset(key="organ", values=["lung"])
    ds.load()
    _ = ds.adata_ls


def test_dsgs_load():
    ds = PrepareData().prepare_dsg(load=False)
    ds.subset(key="organism", values=["Homo sapiens"])
    ds.subset(key="organ", values=["lung"])
    ds.load()


@pytest.mark.parametrize("celltype", ["T cell"])
def test_dsgs_subset_cell_wise(celltype: str):
    """
    Tests if sub-setting results only in datasets of the desired characteristics.
    """
    organ = "lung"
    ds = PrepareData().prepare_dsg(load=False)
    ds.subset(key="organism", values=["Homo sapiens"])
    ds.subset(key="organ", values=[organ])
    ds.load()
    ds.subset_cells(key="cell_type", values=celltype)
    for x in ds.dataset_groups:
        for k, v in x.datasets.items():
            assert v.organism == "Homo sapiens", v.id
            assert v.ontology_container_sfaira.organ.is_a(query=v.organ, reference=organ), v.organ
            for y in np.unique(v.adata.obs[v._adata_ids.cell_type].values):
                assert v.ontology_container_sfaira.cell_type.is_a(query=y, reference=celltype), y


@pytest.mark.parametrize("match_to_release", [RELEASE_HUMAN, {"Homo sapiens": RELEASE_HUMAN}])
@pytest.mark.parametrize("remove_gene_version", [False, True])
@pytest.mark.parametrize("subset_genes_to_type", [None, "protein_coding"])
def test_dsgs_streamline_features(match_to_release: str, remove_gene_version: bool, subset_genes_to_type: str):
    ds = PrepareData().prepare_dsg(load=False)
    ds.subset(key="organism", values=["Homo sapiens"])
    ds.subset(key="organ", values=["lung"])
    ds.load()
    ds.streamline_features(remove_gene_version=remove_gene_version, match_to_release=match_to_release,
                           subset_genes_to_type=subset_genes_to_type)
    gc = sfaira.versions.genomes.GenomeContainer(
        organism="Homo Sapiens",
        release=match_to_release["Homo sapiens"] if isinstance(match_to_release, dict) else match_to_release)
    gc.set(biotype=subset_genes_to_type)
    for x in ds.datasets.values():
        assert x.adata.var["gene_symbol"].tolist() == gc.symbols


def test_dsg_load():
    ds = PrepareData().prepare_dsg(load=False)
    ds.subset(key="organism", values=["Homo sapiens"])
    ds.subset(key="organ", values=["lung"])
    ds = DatasetSuperGroup(dataset_groups=[ds])
    ds.load()


def test_dsg_adata():
    ds = PrepareData().prepare_dsg(load=False)
    ds.subset(key="organism", values=["Homo sapiens"])
    ds.subset(key="organ", values=["lung"])
    ds = DatasetSuperGroup(dataset_groups=[ds])
    _ = ds.adata


def test_ds_interactive():
    adata_ids = AdataIdsSfaira()
    # Prepare object:
    ds = PrepareData().prepare_dsg(load=False)
    ds.subset(key="doi_journal", values=["no_doi_mock1"])
    ds.load()
    adata = ds.adata_ls[0]
    di = DatasetInteractive(data=adata, feature_id_col="index")
    di.organism = "Homo sapiens"
    di.organ = "lung"
    di.cell_type_obs_key = "free_annotation"
    # Test that adata is accessible in non-streamlined object:
    _ = di.adata
    # Test streamlining:
    di.streamline_features(match_to_release=RELEASE_HUMAN)
    di.streamline_metadata(schema="sfaira")
    # Test entries in streamlined object:
    adata_di = di.adata
    assert adata_ids.cell_type in adata_di.obs.columns
    assert adata_ids.cell_type + adata_ids.onto_id_suffix in adata_di.obs.columns
    assert adata_ids.organ in adata_di.uns.keys()
    assert np.all(adata_di.obs[adata_ids.cell_type].values == adata.obs["free_annotation"].values)
    assert adata_di.uns[adata_ids.organ] == "lung"



import anndata
import os
import pandas as pd
import scipy.io

from sfaira.data import DatasetBase


class Dataset(DatasetBase):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.download_url_data = [
            'https://storage.googleapis.com/otar040-cytoimmunogenomics/publication-project-data/scRNAseq/NCOMMS-19-7936188_scRNAseq_raw_UMIs.mtx',
            'https://storage.googleapis.com/otar040-cytoimmunogenomics/publication-project-data/scRNAseq/NCOMMS-19-7936188_scRNAseq_genes.tsv',
            'https://storage.googleapis.com/otar040-cytoimmunogenomics/publication-project-data/scRNAseq/NCOMMS-19-7936188_scRNAseq_barcodes.tsv']
        self.download_url_meta = 'https://storage.googleapis.com/otar040-cytoimmunogenomics/publication-project-data/scRNAseq/NCOMMS-19-7936188_metadata.txt'

        self.assay_sc = "10x 3' v2"
        self.author = "Cano-Gamez"
        self.disease = "healthy"
        self.doi_journal = "10.1038/s41467-020-15543-y"
        self.layer_counts = "X"
        self.organ = "blood"
        self.organism = "Homo sapiens"
        self.primary_data = True
        self.sample_source = "primary_tissue"
        self.year = 2020

        self.feature_symbol_var_key = "names"
        self.feature_type = "rna"
        self.cell_type_obs_key = "cluster.id"

        self.set_dataset_id(idx=1)


def load(data_dir, **kwargs):
    fn = [os.path.join(data_dir, "NCOMMS-19-7936188_scRNAseq_raw_UMIs.mtx"),
          os.path.join(data_dir, 'NCOMMS-19-7936188_scRNAseq_barcodes.tsv'),
          os.path.join(data_dir, 'NCOMMS-19-7936188_scRNAseq_genes.tsv'),
          os.path.join(data_dir, "NCOMMS-19-7936188_metadata.txt")]

    with open(fn[0], 'rb') as mm:
        X = scipy.io.mmread(mm).T.tocsr()
    obs = pd.read_csv(fn[1], header=None, sep="\t", index_col=0)
    obs.index.name = None
    var = pd.read_csv(fn[2], header=None, sep="\t", names=['names'])
    var.index = var['names'].values

    adata = anndata.AnnData(X=X, obs=obs, var=var)

    metadata = pd.read_csv(fn[3], sep="\t")
    adata.obs = adata.obs.join(metadata, how="inner")
    return adata



from sfaira.ui import ModelZoo


def test_for_fatal_embedding():
    model_id = "embedding_homosapiens-lung-linear-0.1-0.1_mylab"
    zoo = ModelZoo()
    zoo.model_id = model_id
    assert zoo.model_id == model_id
    assert zoo.model_class == "embedding"
    assert zoo.model_name == "homosapiens-lung-linear-0.1-0.1"
    assert zoo.organisation == "mylab"
    _ = zoo.topology_container
    _ = zoo.topology_container.topology
    _ = zoo.topology_container.gc


def test_for_fatal_celltype():
    model_id = "celltype_homosapiens-lung-mlp-0.0.1-0.1_mylab"
    zoo = ModelZoo()
    zoo.model_id = model_id
    assert zoo.model_id == model_id
    assert zoo.model_class == "celltype"
    assert zoo.model_name == "homosapiens-lung-mlp-0.0.1-0.1"
    assert zoo.organisation == "mylab"
    _ = zoo.topology_container
    _ = zoo.topology_container.topology
    _ = zoo.topology_container.gc



import os

import anndata
import pytest

from sfaira.unit_tests.data_for_tests.loaders import RELEASE_HUMAN, PrepareData
from sfaira.unit_tests.directories import DIR_TEMP


@pytest.mark.parametrize("out_format", ["sfaira", "cellxgene"])
@pytest.mark.parametrize("clean_obs", [True, False])
@pytest.mark.parametrize("clean_var", [True, False])
@pytest.mark.parametrize("clean_uns", [True, False])
@pytest.mark.parametrize("clean_obs_names", [True, False])
@pytest.mark.parametrize("keep_id_obs", [True])
@pytest.mark.parametrize("keep_orginal_obs", [False])
@pytest.mark.parametrize("keep_symbol_obs", [True])
def test_dsgs_streamline_metadata(out_format: str, clean_obs: bool, clean_var: bool, clean_uns: bool,
                                  clean_obs_names: bool, keep_id_obs: bool, keep_orginal_obs: bool,
                                  keep_symbol_obs: bool):
    ds = PrepareData().prepare_dsg(load=False)
    ds.subset(key="organism", values=["Homo sapiens"])
    ds.subset(key="organ", values=["lung"])
    if out_format == "cellxgene":
        # Other data data sets do not have complete enough annotation
        ds.subset(key="doi_journal", values=["no_doi_mock1", "no_doi_mock3"])
    ds.load()
    ds.streamline_features(remove_gene_version=False, match_to_release=RELEASE_HUMAN,
                           subset_genes_to_type=None)
    ds.streamline_metadata(schema=out_format, clean_obs=clean_obs, clean_var=clean_var,
                           clean_uns=clean_uns, clean_obs_names=clean_obs_names,
                           keep_id_obs=keep_id_obs, keep_orginal_obs=keep_orginal_obs, keep_symbol_obs=keep_symbol_obs)


@pytest.mark.parametrize("schema_version", ["2_0_0"])
@pytest.mark.parametrize("organism", ["Homo sapiens", "Mus musculus"])
def test_cellxgene_export(schema_version: str, organism: str):
    """

    This test can be extended by future versions.
    """
    from cellxgene_schema import validate

    class ValidatorInMemory(validate.Validator):
        """
        Helper class to validate adata in memory and raise errors as in error stream rather than outstream.

        The switch in log stream allows this test to be used as a unit test.
        """

        def validate_adata_inmemory(self, adata: anndata.AnnData):
            self.errors = []
            self.adata = adata
            self._set_schema_def()
            if not self.errors:
                self._deep_check()
            if self.warnings:
                self.warnings = ["WARNING: " + i for i in self.warnings]
            if self.errors:
                self.errors = ["ERROR: " + i for i in self.errors]
            if self.warnings or self.errors:
                print(self.warnings[:20])
                print(self.errors[:20])
                assert False

    ds = PrepareData().prepare_dsg(load=False)
    if organism == "Homo sapiens":
        ds.subset(key="doi_journal", values=["no_doi_mock1"])
    else:
        ds.subset(key="doi_journal", values=["no_doi_mock2"])
    ds.load()
    ds.streamline_features(schema="cellxgene:" + schema_version)
    ds.streamline_metadata(schema="cellxgene:" + schema_version, clean_obs=False, clean_var=True,
                           clean_uns=True, clean_obs_names=False,
                           keep_id_obs=True, keep_orginal_obs=False, keep_symbol_obs=True)
    counter = 0
    for ds in ds.datasets.values():
        # Validate in memory:
        # This directly tests the raw object.
        val = ValidatorInMemory()
        val.validate_adata_inmemory(adata=ds.adata)
        # Validate on disk:
        # This guards against potential issues that arise from h5 formatting.
        fn_temp = os.path.join(DIR_TEMP, "temp.h5ad")
        val = ValidatorInMemory()
        ds.adata.write_h5ad(filename=fn_temp)
        val.validate_adata(h5ad_path=fn_temp)
        counter += 1
    assert counter > 0, "no data sets to test"
